# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A vocabulary that represents the tokens in a dataset and maps them to indices."""

import collections
from typing import Iterable, Optional, Sequence

from absl import logging


class Vocabulary:
  """Represents a vocabulary that can be built from a dataset."""

  def __init__(
      self,
      vocab_path: Optional[str] = None,
      tokenized_sequences: Optional[Iterable[Sequence[bytes]]] = None,
      min_freq: int = 1,
      pad_token: bytes = b'<pad>',
      unk_token: bytes = b'<unk>',
      bos_token: bytes = b'<s>',
      eos_token: bytes = b'</s>',
  ):
    """Loads the vocab from disk (if `vocab_path` is given) or builds it from `tokenized_sequences`."""
    self.pad_token = pad_token
    self.unk_token = unk_token
    self.bos_token = bos_token
    self.eos_token = eos_token
    self.special_tokens = (pad_token, unk_token, bos_token, eos_token)

    if vocab_path:
      self.load(vocab_path)
    elif tokenized_sequences is not None:
      self.build(tokenized_sequences, min_freq=min_freq)
    else:
      raise ValueError(
          (
              'Vocabulary needs either `vocab_path` or `tokenized_sequences` to'
              ' be provided, got %r and %r.'
          )
          % (vocab_path, tokenized_sequences)
      )

  def build(
      self, tokenized_sequences: Iterable[Sequence[bytes]], min_freq: int = 1
  ):
    """Builds a vocabulary over tokens with optional minimum frequency.

    Args:
      tokenized_sequences: Iterable of token sequences to build the vocabulary.
      min_freq: The minimum frequency of each token to be included. Default: 1.
    """
    # Count all the tokens.
    counter = collections.Counter()
    for token_sequence in tokenized_sequences:
      counter.update(token_sequence)

    # Add special tokens to the start of vocab.
    vocab = collections.OrderedDict()
    for token in self.special_tokens:
      vocab[token] = len(vocab)

    # Add all other tokens to the vocab if their frequency is >= min_freq.
    for token, freq in sorted(
        # Sort by frequency (from high to low), and then by token string.
        # This makes sure high frequency tokens get a low token ID.
        counter.items(),
        key=lambda token_freq: (-token_freq[1], token_freq[0]),
    ):
      if freq >= min_freq:
        vocab[token] = len(vocab)

    logging.info('Number of unfiltered tokens: %d', len(counter))
    logging.info('Vocabulary size: %d', len(vocab))
    self.vocab = vocab

  def _getitem__(self, key: str):
    return self.vocab[key]

  def keys(self):
    return self.vocab.keys()

  def values(self):
    return self.vocab.values()

  def __len__(self):
    return len(self.vocab)

  @property
  def pad_idx(self):
    """The index of the padding token."""
    return self.vocab[self.pad_token]

  @property
  def unk_idx(self):
    """The index of the unknown word token."""
    return self.vocab[self.unk_token]

  @property
  def bos_idx(self):
    """The index of the beginning-of-sequence token."""
    return self.vocab[self.bos_token]

  @property
  def eos_idx(self):
    """The index of the end-of-sequence token."""
    return self.vocab[self.eos_token]

  def load(self, path: str) -> None:
    """Loads a vocabulary (one token per line) from disk."""
    vocab = collections.OrderedDict()
    with open(path, 'rb') as f:
      for i, token in enumerate(f):
        assert isinstance(token, bytes), 'Expected byte tokens.'
        vocab[token.rstrip()] = i
    logging.info('Loaded vocabulary, size: %d', len(vocab))
    self.vocab = vocab

  def save(self, path: str) -> None:
    """Saves the vocabulary to disk."""
    with open(path, 'wb') as f:
      for token in self.vocab:
        assert isinstance(token, bytes), 'Expected byte tokens.'
        f.write(token + b'\n')
    logging.info('Saved vocabulary to %s.', path)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from jax import lax
from flax.linen import initializers
from typing import Callable
from flax.linen import Module, compact


class Dense(Module):
  features: int
  kernel_init: Callable = initializers.lecun_normal()
  bias_init: Callable = initializers.zeros_init()
  use_bias: bool = True

  @compact
  def __call__(self, inputs):
    kernel = self.param(
        'kernel', self.kernel_init, (inputs.shape[-1], self.features)
    )
    y = lax.dot_general(
        inputs,
        kernel,
        (((inputs.ndim - 1,), (0,)), ((), ())),
    )
    if self.use_bias:
      bias = self.param('bias', self.bias_init, (self.features,))
      y = y + bias
    return y



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Input pipeline for a LM1B dataset."""

import os
from typing import Dict, Optional, List, Union

from clu import deterministic_data
import ml_collections
import tensorflow as tf
import tensorflow_datasets as tfds

import tokenizer

AUTOTUNE = tf.data.experimental.AUTOTUNE
Features = Dict[str, tf.Tensor]


class NormalizeFeatureNamesOp:
  """Normalizes feature names to 'inputs' and 'targets'."""

  def __init__(self, ds_info: tfds.core.DatasetInfo):
    self.ds_info = ds_info

  def __call__(self, features: Features) -> Features:
    features['inputs'] = features.pop('text')
    # Unnecessary step used for uniformizing with examples/wmt.
    features['targets'] = features['inputs']
    return features


def get_raw_dataset(
    dataset_builder: tfds.core.DatasetBuilder, split: str
) -> tf.data.Dataset:
  """Loads a raw text dataset and normalizes feature keys.

  Args:
    dataset_builder: TFDS dataset builder that can build `split`.
    split: Split to use. This must be the full split. We shard the split across
      multiple hosts and currently don't support sharding subsplits.

  Returns:
    Dataset with source and target language features mapped to 'inputs' and
    'targets'.
  """
  num_examples = dataset_builder.info.splits[split].num_examples
  per_host_split = deterministic_data.get_read_instruction_for_host(
      split, num_examples, drop_remainder=False
  )
  ds = dataset_builder.as_dataset(split=per_host_split, shuffle_files=False)
  ds = ds.map(
      NormalizeFeatureNamesOp(dataset_builder.info), num_parallel_calls=AUTOTUNE
  )
  return ds


def pack_dataset(
    dataset: tf.data.Dataset,
    key2length: Union[int, Dict[str, int]],
    keys: Optional[List[str]] = None,
) -> tf.data.Dataset:
  """Creates a 'packed' version of a dataset on-the-fly.

  Adapted from the mesh-tf implementation.

  This is meant to replace the irritation of having to create a separate
  "packed" version of a dataset to train efficiently on TPU.
  Each example in the output dataset represents several examples in the
  input dataset.
  For each key in the input dataset, two additional keys are created:
  <key>_segmentation: an int32 tensor identifying the parts
     representing the original example.
  <key>_position: an int32 tensor identifying the position within the original
     example.
  Example:
  Two input examples get combined to form an output example.
  The input examples are:
  {"inputs": [8, 7, 1, 0], "targets":[4, 1, 0]}
  {"inputs": [2, 3, 4, 1], "targets":[5, 6, 1]}
  The output example is:
  {
                 "inputs": [8, 7, 1, 2, 3, 4, 1, 0, 0, 0]
    "inputs_segmentation": [1, 1, 1, 2, 2, 2, 2, 0, 0, 0]
        "inputs_position": [0, 1, 2, 0, 1, 2, 3, 0, 0, 0]
                "targets": [4, 1, 5, 6, 1, 0, 0, 0, 0, 0]
   "targets_segmentation": [1, 1, 2, 2, 2, 0, 0, 0, 0, 0]
       "targets_position": [0, 1, 0, 1, 2, 0, 0, 0, 0, 0]
  }
  0 represents padding in both the inputs and the outputs.
  Sequences in the incoming examples are truncated to length "length", and the
  sequences in the output examples all have fixed (padded) length "length".

  Args:
    dataset: a tf.data.Dataset
    key2length: an integer, or a dict from feature-key to integer
    keys: a list of strings (e.g. ["inputs", "targets"])

  Returns:
    a tf.data.Dataset
  """
  shapes = tf.nest.map_structure(lambda spec: spec.shape, dataset.element_spec)
  if keys is None:
    keys = list(shapes.keys())
  for k in keys:
    if k not in shapes:
      raise ValueError(
          'Key %s not found in dataset.  Available keys are %s'
          % (k, shapes.keys())
      )
    if not shapes[k].is_compatible_with(tf.TensorShape([None])):  # type: ignore[wrong-arg-types]
      raise ValueError('Tensors to be packed must be one-dimensional.')
  # make sure that the length dictionary contains all keys as well as the
  # keys suffixed by "_segmentation" and "_position"
  if isinstance(key2length, int):
    key2length = {k: key2length for k in keys}
  for k in keys:
    for suffix in ['_segmentation', '_position']:
      key2length[k + suffix] = key2length[k]

  # trim to length
  dataset = dataset.map(
      lambda x: {k: x[k][: key2length[k]] for k in keys},
      num_parallel_calls=AUTOTUNE,
  )
  # Setting batch_size=length ensures that the concatenated sequences (if they
  # have length >=1) are sufficient to fill at least one packed example.
  batch_size = max(key2length.values())
  dataset = dataset.padded_batch(
      batch_size, padded_shapes={k: [-1] for k in keys}
  )
  dataset = _pack_with_tf_ops(dataset, keys, key2length)

  # Set the Tensor shapes correctly since they get lost in the process.
  def my_fn(x):
    return {k: tf.reshape(v, [key2length[k]]) for k, v in x.items()}

  return dataset.map(my_fn, num_parallel_calls=AUTOTUNE)


def _pack_with_tf_ops(
    dataset: tf.data.Dataset, keys: List[str], key2length: Dict[str, int]
) -> tf.data.Dataset:
  """Helper-function for packing a dataset which has already been batched.

  Helper for pack_dataset()  Uses tf.while_loop.

  Args:
    dataset: a dataset containing padded batches of examples.
    keys: a list of strings
    key2length: a dict from feature-key to integer

  Returns:
    a dataset.
  """
  empty_example = {}
  for k in keys:
    empty_example[k] = tf.zeros([0], dtype=tf.int32)
    empty_example[k + '_position'] = tf.zeros([0], dtype=tf.int32)
  keys_etc = empty_example.keys()

  def write_packed_example(partial, outputs):
    new_partial = empty_example.copy()
    new_outputs = {}
    for k in keys_etc:
      new_outputs[k] = outputs[k].write(
          outputs[k].size(),
          tf.pad(partial[k], [[0, key2length[k] - tf.size(partial[k])]]),
      )
    return new_partial, new_outputs

  def map_fn(x):
    """Internal function to flat_map over.

    Consumes a batch of input examples and produces a variable number of output
    examples.
    Args:
      x: a single example

    Returns:
      a tf.data.Dataset
    """
    partial = empty_example.copy()
    i = tf.zeros([], dtype=tf.int32)
    dynamic_batch_size = tf.shape(x[keys[0]])[0]
    outputs = {}
    for k in keys:
      outputs[k] = tf.TensorArray(
          tf.int32, size=0, dynamic_size=True, element_shape=[key2length[k]]
      )
      outputs[k + '_position'] = tf.TensorArray(
          tf.int32, size=0, dynamic_size=True, element_shape=[key2length[k]]
      )

    def body_fn(i, partial, outputs):
      """Body function for while_loop.

      Args:
        i: integer scalar
        partial: dictionary of Tensor (partially-constructed example)
        outputs: dictionary of TensorArray

      Returns:
        A triple containing the new values of the inputs.
      """
      can_append = True
      one_example = {}
      for k in keys:
        val = tf.cast(x[k][i], tf.int32)
        val = val[: tf.reduce_sum(tf.cast(tf.not_equal(val, 0), tf.int32))]
        one_example[k] = val
      for k in keys:
        can_append = tf.logical_and(
            can_append,
            tf.less_equal(
                tf.size(partial[k]) + tf.size(one_example[k]), key2length[k]
            ),
        )

      def false_fn():
        return write_packed_example(partial, outputs)

      def true_fn():
        return partial, outputs

      partial, outputs = tf.cond(can_append, true_fn, false_fn)
      new_partial = {}
      for k in keys:
        new_seq = one_example[k][: key2length[k]]
        new_seq_len = tf.size(new_seq)
        new_partial[k] = tf.concat([partial[k], new_seq], 0)
        new_partial[k + '_position'] = tf.concat(
            [partial[k + '_position'], tf.range(new_seq_len)], 0
        )
      partial = new_partial
      return i + 1, partial, outputs

    # For loop over all examples in the batch.
    i, partial, outputs = tf.while_loop(
        cond=lambda *_: True,
        body=body_fn,
        loop_vars=(i, partial, outputs),
        shape_invariants=(
            tf.TensorShape([]),
            {k: tf.TensorShape([None]) for k in keys_etc},  # type: ignore[wrong-arg-types]
            {k: tf.TensorShape(None) for k in keys_etc},  # type: ignore[wrong-arg-types]
        ),
        maximum_iterations=dynamic_batch_size,
    )
    _, outputs = write_packed_example(partial, outputs)
    packed = {k: outputs[k].stack() for k in keys_etc}
    for k in keys:
      packed[k + '_segmentation'] = tf.cumsum(
          tf.cast(tf.equal(packed[k + '_position'], 0), tf.int32), axis=1
      ) * tf.cast(tf.not_equal(packed[k], 0), tf.int32)
    return packed

  dataset = dataset.map(map_fn, num_parallel_calls=AUTOTUNE)
  return dataset.unbatch()


# -----------------------------------------------------------------------------
# Main dataset prep routines.
# -----------------------------------------------------------------------------
def preprocess_data(
    dataset,
    shuffle: bool,
    num_epochs: Optional[int] = 1,
    pack_examples: bool = True,
    shuffle_buffer_size: int = 1024,
    max_length: int = 512,
    batch_size: int = 256,
    drop_remainder: bool = True,
    prefetch_size: int = AUTOTUNE,
):
  """Shuffle and batch/pack the given dataset."""

  def length_filter(max_len):
    def filter_fn(x):
      source, target = x['inputs'], x['targets']
      l = tf.maximum(tf.shape(source)[0], tf.shape(target)[0])
      return tf.less(l, max_len + 1)

    return filter_fn

  if max_length > 0:
    dataset = dataset.filter(length_filter(max_length))

  if shuffle:
    dataset = dataset.shuffle(shuffle_buffer_size)
  dataset = dataset.repeat(num_epochs)

  if pack_examples:
    dataset = pack_dataset(dataset, max_length)
    dataset = dataset.batch(batch_size, drop_remainder=drop_remainder)
  else:  # simple (static-shape) padded batching
    dataset = dataset.padded_batch(
        batch_size,
        padded_shapes={'inputs': max_length, 'targets': max_length},
        padding_values={'inputs': 0, 'targets': 0},
        drop_remainder=drop_remainder,
    )

  if prefetch_size:
    dataset = dataset.prefetch(prefetch_size)

  return dataset


def get_datasets(
    config: ml_collections.ConfigDict,
    *,
    n_devices: int,
    vocab_path: Optional[str] = None,
):
  """Load and return dataset of batched examples for use during training."""
  if vocab_path is None:
    vocab_path = os.path.expanduser('~/lm1b_sentencepiece_model')

  train_ds_builder = tfds.builder(config.dataset_name)
  train_data = get_raw_dataset(train_ds_builder, 'train')

  if config.eval_dataset_name:
    eval_ds_builder = tfds.builder(config.eval_dataset_name)
  else:
    eval_ds_builder = train_ds_builder
  eval_data = get_raw_dataset(eval_ds_builder, config.eval_split)

  # Tokenize data.
  sp_tokenizer = tokenizer.load_or_train_tokenizer(
      train_data,
      vocab_path=vocab_path,
      vocab_size=config.vocab_size,
      max_corpus_chars=config.max_corpus_chars,
  )
  train_data = train_data.map(
      tokenizer.TokenizeOp(sp_tokenizer), num_parallel_calls=AUTOTUNE
  )
  eval_data = eval_data.map(
      tokenizer.TokenizeOp(sp_tokenizer), num_parallel_calls=AUTOTUNE
  )

  batch_size = config.per_device_batch_size * n_devices
  if config.eval_per_device_batch_size > 0:
    eval_batch_size = config.eval_per_device_batch_size * n_devices
  else:
    eval_batch_size = batch_size

  train_ds = preprocess_data(
      train_data,
      shuffle=True,
      num_epochs=None,
      pack_examples=True,
      batch_size=batch_size,
      max_length=config.max_target_length,
  )

  eval_ds = preprocess_data(
      eval_data,
      shuffle=False,
      pack_examples=False,
      batch_size=eval_batch_size,
      max_length=config.max_eval_target_length,
  )

  predict_ds = preprocess_data(
      eval_data,
      shuffle=False,
      pack_examples=False,
      batch_size=eval_batch_size,
      max_length=config.max_predict_length,
      drop_remainder=False,
  )

  return train_ds, eval_ds, predict_ds, sp_tokenizer



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import typing as tp
from typing import Any

from flax import linen
from flax.experimental.nnx.nnx import helpers
from flax.experimental.nnx.nnx import variables as variableslib
from flax.experimental.nnx.nnx.module import GraphDef, Module
from flax.experimental.nnx.nnx.rnglib import Rngs
from flax.experimental.nnx.nnx.state import State

M = tp.TypeVar('M', bound=Module)


# Flax-like style is NNX
@dataclasses.dataclass
class Functional(tp.Generic[M]):
  module_type: tp.Type[M]
  graphdef: tp.Optional[GraphDef[M]]
  args: tuple[tp.Any, ...]
  kwargs: dict[str, tp.Any]

  def init(self, *, rngs: tp.Optional[Rngs] = None) -> State:
    kwargs = {}
    if rngs is not None:
      kwargs['rngs'] = rngs
    module = self.module_type(*self.args, **self.kwargs, **kwargs)
    state, graphdef = module.split()
    self.graphdef = graphdef
    return state

  def apply(self, *states: tp.Any):
    assert self.graphdef is not None
    return self.graphdef.apply(*states)


def functional(cls: tp.Type[M]) -> tp.Callable[..., Functional[M]]:
  def _functional_constructor(*args: tp.Any, **kwargs: tp.Any) -> Functional[M]:
    return Functional(cls, None, args, kwargs)

  return _functional_constructor


class LinenWrapper(Module):
  def __init__(
    self,
    module: linen.Module,
    *args: tp.Any,
    rngs: tp.Optional[Rngs] = None,
    **kwargs: tp.Any,
  ):
    self.module = module

    _rngs = (
      {name: stream.key for name, stream in rngs._rngs.items()} if rngs else {}
    )
    # rename default to params
    if 'params' not in _rngs and 'default' in _rngs:
      _rngs['params'] = _rngs['default']
      del _rngs['default']

    variables = module.init(_rngs, *args, **kwargs)

    self.states = helpers.Dict(
      (collection, variableslib.variable_type(collection)(value))
      for collection, value in variables.items()
    )

  def __call__(
    self, *args: Any, rngs: tp.Optional[Rngs] = None, **kwargs: Any
  ) -> Any:
    _rngs = (
      {name: stream.key for name, stream in rngs._rngs.items()} if rngs else {}
    )

    variables = {collection: value for collection, value in self.states.items()}
    out = self.module.apply(variables, *args, rngs=_rngs, **kwargs)

    if kwargs.get('mutable', False) != False:
      out, updates = out
      for collection, value in updates.items():
        if collection in self.states:
          self.states[collection] = value
        else:
          self.states[collection] = variableslib.variable_type(collection)(
            value
          )

    return out


class NNXWrapper(linen.Module):
  ...



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import enum
from typing import (
  Any,
  Callable,
  Dict,
  Generator,
  Generic,
  Mapping,
  Optional,
  Protocol,
  TypeVar,
  runtime_checkable,
)

from flax.core import FrozenDict
from flax.errors import CursorFindError, TraverseTreeError

A = TypeVar('A')
Key = Any


@runtime_checkable
class Indexable(Protocol):
  def __getitem__(self, key) -> Any:
    ...


class AccessType(enum.Enum):
  ITEM = enum.auto()
  ATTR = enum.auto()


@dataclasses.dataclass
class ParentKey(Generic[A]):
  parent: 'Cursor[A]'
  key: Any
  access_type: AccessType


def is_named_tuple(obj):
  return (
    isinstance(obj, tuple)
    and hasattr(obj, '_fields')
    and hasattr(obj, '_asdict')
    and hasattr(obj, '_replace')
  )


def _traverse_tree(path, obj, *, update_fn=None, cond_fn=None):
  """Helper function for ``Cursor.apply_update`` and ``Cursor.find_all``.
  Exactly one of ``update_fn`` and ``cond_fn`` must be not None.

  - If ``update_fn`` is not None, then ``Cursor.apply_update`` is calling
    this function and ``_traverse_tree`` will return a generator where
    each generated element is of type Tuple[Tuple[Union[str, int], AccessType], Any].
    The first element is a tuple of the key path and access type where the
    change was applied from the ``update_fn``, and the second element is
    the newly modified value. If the generator is non-empty, then the
    tuple key path will always be non-empty as well.
  - If ``cond_fn`` is not None, then ``Cursor.find_all`` is calling this
    function and ``_traverse_tree`` will return a generator where each
    generated element is of type Tuple[Union[str, int], AccessType]. The
    tuple contains the key path and access type where the object was found
    that fulfilled the conditions of the ``cond_fn``.
  """
  if not (bool(update_fn) ^ bool(cond_fn)):
    raise TraverseTreeError(update_fn, cond_fn)

  if path:
    str_path = '/'.join(str(key) for key, _ in path)
    if update_fn:
      new_obj = update_fn(str_path, obj)
      if new_obj is not obj:
        yield path, new_obj
        return
    elif cond_fn(str_path, obj):  # type: ignore
      yield path
      return

  if isinstance(obj, (FrozenDict, dict)):
    items = obj.items()
    access_type = AccessType.ITEM
  elif is_named_tuple(obj):
    items = ((name, getattr(obj, name)) for name in obj._fields)  # type: ignore
    access_type = AccessType.ATTR
  elif isinstance(obj, (list, tuple)):
    items = enumerate(obj)
    access_type = AccessType.ITEM
  elif dataclasses.is_dataclass(obj):
    items = (
      (f.name, getattr(obj, f.name)) for f in dataclasses.fields(obj) if f.init
    )
    access_type = AccessType.ATTR
  else:
    return

  if update_fn:
    for key, value in items:
      yield from _traverse_tree(
        path + ((key, access_type),), value, update_fn=update_fn
      )
  else:
    for key, value in items:
      yield from _traverse_tree(
        path + ((key, access_type),), value, cond_fn=cond_fn
      )


class Cursor(Generic[A]):
  _obj: A
  _parent_key: Optional[ParentKey[A]]
  _changes: Dict[Any, 'Cursor[A]']

  def __init__(self, obj: A, parent_key: Optional[ParentKey[A]]):
    # NOTE: we use `vars` here to avoid calling `__setattr__`
    # vars(self) = self.__dict__
    vars(self)['_obj'] = obj
    vars(self)['_parent_key'] = parent_key
    vars(self)['_changes'] = {}

  @property
  def _root(self) -> 'Cursor[A]':
    if self._parent_key is None:
      return self
    else:
      return self._parent_key.parent._root  # type: ignore

  @property
  def _path(self) -> str:
    if self._parent_key is None:
      return ''
    if self._parent_key.access_type == AccessType.ITEM:  # type: ignore
      if isinstance(self._parent_key.key, str):  # type: ignore
        key = "'" + self._parent_key.key + "'"  # type: ignore
      else:
        key = str(self._parent_key.key)  # type: ignore
      return self._parent_key.parent._path + '[' + key + ']'  # type: ignore
    # self.parent_key.access_type == AccessType.ATTR:
    return self._parent_key.parent._path + '.' + self._parent_key.key  # type: ignore

  def __getitem__(self, key) -> 'Cursor[A]':
    if key in self._changes:
      return self._changes[key]

    if not isinstance(self._obj, Indexable):
      raise TypeError(f'Cannot index into {self._obj}')

    if isinstance(self._obj, Mapping) and key not in self._obj:
      raise KeyError(f'Key {key} not found in {self._obj}')

    if is_named_tuple(self._obj):
      return getattr(self, self._obj._fields[key])  # type: ignore

    child = Cursor(self._obj[key], ParentKey(self, key, AccessType.ITEM))
    self._changes[key] = child
    return child

  def __getattr__(self, name) -> 'Cursor[A]':
    if name in self._changes:
      return self._changes[name]

    if not hasattr(self._obj, name):
      raise AttributeError(f'Attribute {name} not found in {self._obj}')

    child = Cursor(
      getattr(self._obj, name), ParentKey(self, name, AccessType.ATTR)
    )
    self._changes[name] = child
    return child

  def __setitem__(self, key, value):
    if is_named_tuple(self._obj):
      return setattr(self, self._obj._fields[key], value)  # type: ignore
    self._changes[key] = Cursor(value, ParentKey(self, key, AccessType.ITEM))

  def __setattr__(self, name, value):
    self._changes[name] = Cursor(value, ParentKey(self, name, AccessType.ATTR))

  def set(self, value) -> A:
    """Set a new value for an attribute, property, element or entry
    in the Cursor object and return a copy of the original object,
    containing the new set value.

    Example::

      >>> from flax.cursor import cursor
      >>> from flax.training import train_state
      >>> import optax

      >>> dict_obj = {'a': 1, 'b': (2, 3), 'c': [4, 5]}
      >>> modified_dict_obj = cursor(dict_obj)['b'][0].set(10)
      >>> assert modified_dict_obj == {'a': 1, 'b': (10, 3), 'c': [4, 5]}

      >>> state = train_state.TrainState.create(
      ...     apply_fn=lambda x: x,
      ...     params=dict_obj,
      ...     tx=optax.adam(1e-3),
      ... )
      >>> modified_state = cursor(state).params['b'][1].set(10)
      >>> assert modified_state.params == {'a': 1, 'b': (2, 10), 'c': [4, 5]}

    Args:
      value: the value used to set an attribute, property, element or entry in the Cursor object
    Returns:
      A copy of the original object with the new set value.
    """
    if self._parent_key is None:
      return value
    parent, key = self._parent_key.parent, self._parent_key.key  # type: ignore
    parent._changes[key] = Cursor(value, self._parent_key)
    return parent._root.build()

  def build(self) -> A:
    """Create and return a copy of the original object with accumulated changes.
    This method is to be called after making changes to the Cursor object.

    NOTE: The new object is built bottom-up, the changes will be first applied
    to the leaf nodes, and then its parent, all the way up to the root.

    Example::

      >>> from flax.cursor import cursor
      >>> from flax.training import train_state
      >>> import optax

      >>> dict_obj = {'a': 1, 'b': (2, 3), 'c': [4, 5]}
      >>> c = cursor(dict_obj)
      >>> c['b'][0] = 10
      >>> c['a'] = (100, 200)
      >>> modified_dict_obj = c.build()
      >>> assert modified_dict_obj == {'a': (100, 200), 'b': (10, 3), 'c': [4, 5]}

      >>> state = train_state.TrainState.create(
      ...     apply_fn=lambda x: x,
      ...     params=dict_obj,
      ...     tx=optax.adam(1e-3),
      ... )
      >>> new_fn = lambda x: x + 1
      >>> c = cursor(state)
      >>> c.params['b'][1] = 10
      >>> c.apply_fn = new_fn
      >>> modified_state = c.build()
      >>> assert modified_state.params == {'a': 1, 'b': (2, 10), 'c': [4, 5]}
      >>> assert modified_state.apply_fn == new_fn

    Returns:
      A copy of the original object with the accumulated changes.
    """
    changes = {
      key: child.build() if isinstance(child, Cursor) else child
      for key, child in self._changes.items()
    }
    if isinstance(self._obj, FrozenDict):
      obj = self._obj.copy(changes)  # type: ignore
    elif isinstance(self._obj, (dict, list)):
      obj = self._obj.copy()  # type: ignore
      for key, value in changes.items():
        obj[key] = value
    elif is_named_tuple(self._obj):
      obj = self._obj._replace(**changes)  # type: ignore
    elif isinstance(self._obj, tuple):
      obj = list(self._obj)  # type: ignore
      for key, value in changes.items():
        obj[key] = value
      obj = tuple(obj)  # type: ignore
    elif dataclasses.is_dataclass(self._obj):
      obj = dataclasses.replace(self._obj, **changes)  # type: ignore
    else:
      obj = self._obj  # type: ignore
    return obj  # type: ignore

  def apply_update(
    self,
    update_fn: Callable[[str, Any], Any],
  ) -> 'Cursor[A]':
    """Traverse the Cursor object and record conditional changes recursively via an ``update_fn``.
    The changes are recorded in the Cursor object's ``._changes`` dictionary. To generate a copy
    of the original object with the accumulated changes, call the ``.build`` method after calling
    ``.apply_update``.

    The ``update_fn`` has a function signature of ``(str, Any) -> Any``:

    - The input arguments are the current key path (in the form of a string delimited
      by ``'/'``) and value at that current key path
    - The output is the new value (either modified by the ``update_fn`` or same as the
      input value if the condition wasn't fulfilled)

    NOTES:

    - If the ``update_fn`` returns a modified value, this method will not recurse any further
      down that branch to record changes. For example, if we intend to replace an attribute that points
      to a dictionary with an int, we don't need to look for further changes inside the dictionary,
      since the dictionary will be replaced anyways.
    - The ``is`` operator is used to determine whether the return value is modified (by comparing it
      to the input value). Therefore if the ``update_fn`` modifies a mutable container (e.g. lists,
      dicts, etc.) and returns the same container, ``.apply_update`` will treat the returned value as
      unmodified as it contains the same ``id``. To avoid this, return a copy of the modified value.
    - ``.apply_update`` WILL NOT call the ``update_fn`` to the value at the top-most level of
      the pytree (i.e. the root node). The ``update_fn`` will first be called on the root node's
      children, and then the pytree traversal will continue recursively from there.

    Example::

      >>> import flax.linen as nn
      >>> from flax.cursor import cursor
      >>> import jax, jax.numpy as jnp

      >>> class Model(nn.Module):
      ...   @nn.compact
      ...   def __call__(self, x):
      ...     x = nn.Dense(3)(x)
      ...     x = nn.relu(x)
      ...     x = nn.Dense(3)(x)
      ...     x = nn.relu(x)
      ...     x = nn.Dense(3)(x)
      ...     x = nn.relu(x)
      ...     return x

      >>> params = Model().init(jax.random.key(0), jnp.empty((1, 2)))['params']

      >>> def update_fn(path, value):
      ...   '''Multiply all dense kernel params by 2 and add 1.
      ...   Subtract the Dense_1 bias param by 1.'''
      ...   if 'kernel' in path:
      ...     return value * 2 + 1
      ...   elif 'Dense_1' in path and 'bias' in path:
      ...     return value - 1
      ...   return value

      >>> c = cursor(params)
      >>> new_params = c.apply_update(update_fn).build()
      >>> for layer in ('Dense_0', 'Dense_1', 'Dense_2'):
      ...   assert (new_params[layer]['kernel'] == 2 * params[layer]['kernel'] + 1).all()
      ...   if layer == 'Dense_1':
      ...     assert (new_params[layer]['bias'] == params[layer]['bias'] - 1).all()
      ...   else:
      ...     assert (new_params[layer]['bias'] == params[layer]['bias']).all()

      >>> assert jax.tree_util.tree_all(
      ...       jax.tree_util.tree_map(
      ...           lambda x, y: (x == y).all(),
      ...           params,
      ...           Model().init(jax.random.key(0), jnp.empty((1, 2)))[
      ...               'params'
      ...           ],
      ...       )
      ...   ) # make sure original params are unchanged

    Args:
      update_fn: the function that will conditionally record changes to the Cursor object
    Returns:
      The current Cursor object with the recorded conditional changes specified by the
      ``update_fn``. To generate a copy of the original object with the accumulated
      changes, call the ``.build`` method after calling ``.apply_update``.
    """
    for path, value in _traverse_tree((), self._obj, update_fn=update_fn):
      child = self
      for key, access_type in path[:-1]:
        if access_type is AccessType.ITEM:
          child = child[key]
        else:  # access_type is AccessType.ATTR
          child = getattr(child, key)
      key, access_type = path[-1]
      if access_type is AccessType.ITEM:
        child[key] = value
      else:  # access_type is AccessType.ATTR
        setattr(child, key, value)

    return self

  def find(self, cond_fn: Callable[[str, Any], bool]) -> 'Cursor[A]':
    """Traverse the Cursor object and return a child Cursor object that fulfill the
    conditions in the ``cond_fn``. The ``cond_fn`` has a function signature of ``(str, Any) -> bool``:

    - The input arguments are the current key path (in the form of a string delimited
      by ``'/'``) and value at that current key path
    - The output is a boolean, denoting whether to return the child Cursor object at this path

    Raises a :meth:`CursorFindError <flax.errors.CursorFindError>` if no object or more
    than one object is found that fulfills the condition of the ``cond_fn``. We raise an
    error because the user should always expect this method to return the only object whose
    corresponding key path and value fulfill the condition of the ``cond_fn``.

    NOTES:

    - If the ``cond_fn`` evaluates to True at a particular key path, this method will not recurse
      any further down that branch; i.e. this method will find and return the "earliest" child node
      that fulfills the condition in ``cond_fn`` in a particular key path
    - ``.find`` WILL NOT search the the value at the top-most level of the pytree (i.e. the root
      node). The ``cond_fn`` will be evaluated recursively, starting at the root node's children.

    Example::

      >>> import flax.linen as nn
      >>> from flax.cursor import cursor
      >>> import jax, jax.numpy as jnp

      >>> class Model(nn.Module):
      ...   @nn.compact
      ...   def __call__(self, x):
      ...     x = nn.Dense(3)(x)
      ...     x = nn.relu(x)
      ...     x = nn.Dense(3)(x)
      ...     x = nn.relu(x)
      ...     x = nn.Dense(3)(x)
      ...     x = nn.relu(x)
      ...     return x

      >>> params = Model().init(jax.random.PRNGKey(0), jnp.empty((1, 2)))['params']

      >>> def cond_fn(path, value):
      ...   '''Find the second dense layer params.'''
      ...   return 'Dense_1' in path

      >>> new_params = cursor(params).find(cond_fn)['bias'].set(params['Dense_1']['bias'] + 1)

      >>> for layer in ('Dense_0', 'Dense_1', 'Dense_2'):
      ...   if layer == 'Dense_1':
      ...     assert (new_params[layer]['bias'] == params[layer]['bias'] + 1).all()
      ...   else:
      ...     assert (new_params[layer]['bias'] == params[layer]['bias']).all()

      >>> c = cursor(params)
      >>> c2 = c.find(cond_fn)
      >>> c2['kernel'] += 2
      >>> c2['bias'] += 2
      >>> new_params = c.build()

      >>> for layer in ('Dense_0', 'Dense_1', 'Dense_2'):
      ...   if layer == 'Dense_1':
      ...     assert (new_params[layer]['kernel'] == params[layer]['kernel'] + 2).all()
      ...     assert (new_params[layer]['bias'] == params[layer]['bias'] + 2).all()
      ...   else:
      ...     assert (new_params[layer]['kernel'] == params[layer]['kernel']).all()
      ...     assert (new_params[layer]['bias'] == params[layer]['bias']).all()

      >>> assert jax.tree_util.tree_all(
      ...       jax.tree_util.tree_map(
      ...           lambda x, y: (x == y).all(),
      ...           params,
      ...           Model().init(jax.random.PRNGKey(0), jnp.empty((1, 2)))[
      ...               'params'
      ...           ],
      ...       )
      ...   ) # make sure original params are unchanged

    Args:
      cond_fn: the function that will conditionally find child Cursor objects
    Returns:
      A child Cursor object that fulfills the condition in the ``cond_fn``.
    """
    generator = self.find_all(cond_fn)
    try:
      cursor = next(generator)
    except StopIteration:
      raise CursorFindError()
    try:
      cursor2 = next(generator)
      raise CursorFindError(cursor, cursor2)
    except StopIteration:
      return cursor

  def find_all(
    self, cond_fn: Callable[[str, Any], bool]
  ) -> Generator['Cursor[A]', None, None]:
    """Traverse the Cursor object and return a generator of child Cursor objects that fulfill the
    conditions in the ``cond_fn``. The ``cond_fn`` has a function signature of ``(str, Any) -> bool``:

    - The input arguments are the current key path (in the form of a string delimited
      by ``'/'``) and value at that current key path
    - The output is a boolean, denoting whether to return the child Cursor object at this path

    NOTES:

    - If the ``cond_fn`` evaluates to True at a particular key path, this method will not recurse
      any further down that branch; i.e. this method will find and return the "earliest" child nodes
      that fulfill the condition in ``cond_fn`` in a particular key path
    - ``.find_all`` WILL NOT search the the value at the top-most level of the pytree (i.e. the root
      node). The ``cond_fn`` will be evaluated recursively, starting at the root node's children.

    Example::

      >>> import flax.linen as nn
      >>> from flax.cursor import cursor
      >>> import jax, jax.numpy as jnp

      >>> class Model(nn.Module):
      ...   @nn.compact
      ...   def __call__(self, x):
      ...     x = nn.Dense(3)(x)
      ...     x = nn.relu(x)
      ...     x = nn.Dense(3)(x)
      ...     x = nn.relu(x)
      ...     x = nn.Dense(3)(x)
      ...     x = nn.relu(x)
      ...     return x

      >>> params = Model().init(jax.random.PRNGKey(0), jnp.empty((1, 2)))['params']

      >>> def cond_fn(path, value):
      ...   '''Find all dense layer params.'''
      ...   return 'Dense' in path

      >>> c = cursor(params)
      >>> for dense_params in c.find_all(cond_fn):
      ...   dense_params['bias'] += 1
      >>> new_params = c.build()

      >>> for layer in ('Dense_0', 'Dense_1', 'Dense_2'):
      ...   assert (new_params[layer]['bias'] == params[layer]['bias'] + 1).all()

      >>> assert jax.tree_util.tree_all(
      ...       jax.tree_util.tree_map(
      ...           lambda x, y: (x == y).all(),
      ...           params,
      ...           Model().init(jax.random.PRNGKey(0), jnp.empty((1, 2)))[
      ...               'params'
      ...           ],
      ...       )
      ...   ) # make sure original params are unchanged

    Args:
      cond_fn: the function that will conditionally find child Cursor objects
    Returns:
      A generator of child Cursor objects that fulfill the condition in the ``cond_fn``.
    """
    for path in _traverse_tree((), self._obj, cond_fn=cond_fn):
      child = self
      for key, access_type in path:
        if access_type is AccessType.ITEM:
          child = child[key]
        else:  # access_type is AccessType.ATTR
          child = getattr(child, key)
      yield child

  def __str__(self):
    return str(self._obj)

  def __repr__(self):
    return self._pretty_repr()

  def _pretty_repr(self, indent=2, _prefix_indent=0):
    s = 'Cursor(\n'
    obj_str = repr(self._obj).replace(
      '\n', '\n' + ' ' * (_prefix_indent + indent)
    )
    s += ' ' * (_prefix_indent + indent) + f'_obj={obj_str},\n'
    s += ' ' * (_prefix_indent + indent) + '_changes={'
    if self._changes:
      s += '\n'
      for key in self._changes:
        str_key = repr(key)
        prefix = ' ' * (_prefix_indent + 2 * indent) + str_key + ': '
        s += (
          prefix
          + self._changes[key]._pretty_repr(
            indent=indent, _prefix_indent=len(prefix)
          )
          + ',\n'
        )
      s = s[
        :-2
      ]  # remove comma and newline character for last element in self._changes
      s += '\n' + ' ' * (_prefix_indent + indent) + '}\n'
    else:
      s += '}\n'
    s += ' ' * _prefix_indent + ')'
    return s

  def __len__(self):
    return len(self._obj)

  def __iter__(self):
    if isinstance(self._obj, (tuple, list)):
      return (self[i] for i in range(len(self._obj)))
    else:
      raise NotImplementedError(
        '__iter__ method only implemented for tuples and lists, not type'
        f' {type(self._obj)}'
      )

  def __reversed__(self):
    if isinstance(self._obj, (tuple, list)):
      return (self[i] for i in range(len(self._obj) - 1, -1, -1))
    else:
      raise NotImplementedError(
        '__reversed__ method only implemented for tuples and lists, not type'
        f' {type(self._obj)}'
      )

  def __add__(self, other):
    return self._obj + other

  def __sub__(self, other):
    return self._obj - other

  def __mul__(self, other):
    return self._obj * other

  def __matmul__(self, other):
    return self._obj @ other

  def __truediv__(self, other):
    return self._obj / other

  def __floordiv__(self, other):
    return self._obj // other

  def __mod__(self, other):
    return self._obj % other

  def __divmod__(self, other):
    return divmod(self._obj, other)

  def __pow__(self, other):
    return pow(self._obj, other)

  def __lshift__(self, other):
    return self._obj << other

  def __rshift__(self, other):
    return self._obj >> other

  def __and__(self, other):
    return self._obj & other

  def __xor__(self, other):
    return self._obj ^ other

  def __or__(self, other):
    return self._obj | other

  def __radd__(self, other):
    return other + self._obj

  def __rsub__(self, other):
    return other - self._obj

  def __rmul__(self, other):
    return other * self._obj

  def __rmatmul__(self, other):
    return other @ self._obj

  def __rtruediv__(self, other):
    return other / self._obj

  def __rfloordiv__(self, other):
    return other // self._obj

  def __rmod__(self, other):
    return other % self._obj

  def __rdivmod__(self, other):
    return divmod(other, self._obj)

  def __rpow__(self, other):
    return pow(other, self._obj)

  def __rlshift__(self, other):
    return other << self._obj

  def __rrshift__(self, other):
    return other >> self._obj

  def __rand__(self, other):
    return other & self._obj

  def __rxor__(self, other):
    return other ^ self._obj

  def __ror__(self, other):
    return other | self._obj

  def __neg__(self):
    return -self._obj

  def __pos__(self):
    return +self._obj

  def __abs__(self):
    return abs(self._obj)

  def __invert__(self):
    return ~self._obj

  def __round__(self, ndigits=None):
    return round(self._obj, ndigits)

  def __lt__(self, other):
    return self._obj < other

  def __le__(self, other):
    return self._obj <= other

  def __eq__(self, other):
    return self._obj == other

  def __ne__(self, other):
    return self._obj != other

  def __gt__(self, other):
    return self._obj > other

  def __ge__(self, other):
    return self._obj >= other


def cursor(obj: A) -> Cursor[A]:
  """Wrap :class:`Cursor <flax.cursor.Cursor>` over ``obj`` and return it.
  Changes can then be applied to the Cursor object in the following ways:

  - single-line change via the ``.set`` method
  - multiple changes, and then calling the ``.build`` method
  - multiple changes conditioned on the pytree path and node value via the
    ``.apply_update`` method, and then calling the ``.build`` method

  ``.set`` example::

    >>> from flax.cursor import cursor

    >>> dict_obj = {'a': 1, 'b': (2, 3), 'c': [4, 5]}
    >>> modified_dict_obj = cursor(dict_obj)['b'][0].set(10)
    >>> assert modified_dict_obj == {'a': 1, 'b': (10, 3), 'c': [4, 5]}

  ``.build`` example::

    >>> from flax.cursor import cursor

    >>> dict_obj = {'a': 1, 'b': (2, 3), 'c': [4, 5]}
    >>> c = cursor(dict_obj)
    >>> c['b'][0] = 10
    >>> c['a'] = (100, 200)
    >>> modified_dict_obj = c.build()
    >>> assert modified_dict_obj == {'a': (100, 200), 'b': (10, 3), 'c': [4, 5]}

  ``.apply_update`` example::

    >>> from flax.cursor import cursor
    >>> from flax.training import train_state
    >>> import optax

    >>> def update_fn(path, value):
    ...   '''Replace params with empty dictionary.'''
    ...   if 'params' in path:
    ...     return {}
    ...   return value

    >>> state = train_state.TrainState.create(
    ...     apply_fn=lambda x: x,
    ...     params={'a': 1, 'b': 2},
    ...     tx=optax.adam(1e-3),
    ... )
    >>> c = cursor(state)
    >>> state2 = c.apply_update(update_fn).build()
    >>> assert state2.params == {}
    >>> assert state.params == {'a': 1, 'b': 2} # make sure original params are unchanged

  If the underlying ``obj`` is a ``list`` or ``tuple``, iterating over the Cursor object
  to get the child Cursors is also possible::

    >>> from flax.cursor import cursor

    >>> c = cursor(((1, 2), (3, 4)))
    >>> for child_c in c:
    ...   child_c[1] *= -1
    >>> assert c.build() == ((1, -2), (3, -4))

  View the docstrings for each method to see more examples of their usage.

  Args:
    obj: the object you want to wrap the Cursor in
  Returns:
    A Cursor object wrapped around obj.
  """
  return Cursor(obj, None)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Flax Module summary library."""

import dataclasses
import enum
import io
from abc import ABC, abstractmethod
from types import MappingProxyType
from typing import (
  Any,
  Callable,
  Dict,
  Iterable,
  List,
  Mapping,
  Optional,
  Sequence,
  Set,
  Tuple,
  Type,
  Union,
)

import jax
import jax.numpy as jnp
import numpy as np
import rich.console
import rich.table
import rich.text
import yaml

import flax.linen.module as module_lib
from flax.core import meta, unfreeze
from flax.core.scope import (
  CollectionFilter,
  DenyList,
  FrozenVariableDict,
  LazyRng,
  MutableVariableDict,
)

PRNGKey = Any  # pylint: disable=invalid-name
RNGSequences = Dict[str, PRNGKey]
Array = Any  # pylint: disable=invalid-name


class _ValueRepresentation(ABC):
  """A class that represents a value in the summary table."""

  @abstractmethod
  def render(self) -> str:
    ...


@dataclasses.dataclass
class _ArrayRepresentation(_ValueRepresentation):
  shape: Tuple[int, ...]
  dtype: Any

  @classmethod
  def from_array(cls, x: Array) -> '_ArrayRepresentation':
    return cls(jnp.shape(x), jnp.result_type(x))

  @classmethod
  def render_array(cls, x) -> str:
    return cls.from_array(x).render()

  def render(self):
    shape_repr = ','.join(str(x) for x in self.shape)
    return f'[dim]{self.dtype}[/dim][{shape_repr}]'


@dataclasses.dataclass
class _PartitionedArrayRepresentation(_ValueRepresentation):
  array_representation: _ArrayRepresentation
  names: meta.LogicalNames

  @classmethod
  def from_partitioned(
    cls, partitioned: meta.Partitioned
  ) -> '_PartitionedArrayRepresentation':
    return cls(
      _ArrayRepresentation.from_array(partitioned.value), partitioned.names
    )

  def render(self):
    return self.array_representation.render() + f' [dim]P[/dim]{self.names}'


@dataclasses.dataclass
class _ObjectRepresentation(_ValueRepresentation):
  obj: Any

  def render(self):
    return repr(self.obj)


@dataclasses.dataclass
class Row:
  """Contains the information about a single row in the summary table.

  Attributes:
    path: A tuple of strings that represents the path to the module.
    module_type: type of the Module.
    method: method of the module called.
    inputs: inputs to the module.
    outputs: Output of the Module as reported by `capture_intermediates`.
    module_variables: Dictionary of variables in the module (no submodules
      included).
    counted_variables: Dictionary of variables that should be counted for this
      row, if no summarization is done (e.g. `depth=None` in `module_summary`)
      then this field is the same as `module_variables`, however if a
      summarization is done then this dictionary potentially contains parameters
      from submodules depending on the depth of the Module in question.
    flops: FLOPs cost of calling the module method.
    vjp_flops: FLOPs cost of calling the VJP of the module method.
  """

  path: Tuple[str, ...]
  module_type: Type[module_lib.Module]
  method: str
  inputs: Any
  outputs: Any
  module_variables: Dict[str, Dict[str, Any]]
  counted_variables: Dict[str, Dict[str, Any]]
  flops: int
  vjp_flops: int

  def __post_init__(self):
    self.inputs = self.inputs
    self.outputs = self.outputs
    self.module_variables = self.module_variables
    self.counted_variables = self.counted_variables

  def size_and_bytes(
    self, collections: Iterable[str]
  ) -> Dict[str, Tuple[int, int]]:
    return {
      col: (
        _size_and_bytes(self.counted_variables[col])
        if col in self.counted_variables
        else (0, 0)
      )
      for col in collections
    }


class Table(List[Row]):
  """A list of Row objects.

  Table inherits from `List[Row]` so it has all the methods of a list, however
  it also contains some additional fields:

  * `module`: the module that this table is summarizing
  * `collections`: a list containing the parameter collections (e.g. 'params', 'batch_stats', etc)
  """

  def __init__(
    self,
    module: module_lib.Module,
    collections: Sequence[str],
    rows: Iterable[Row],
  ):
    super().__init__(rows)
    self.module = module
    self.collections = collections


def tabulate(
  module: module_lib.Module,
  rngs: Union[PRNGKey, RNGSequences],
  depth: Optional[int] = None,
  show_repeated: bool = False,
  mutable: CollectionFilter = DenyList('intermediates'),
  console_kwargs: Optional[Mapping[str, Any]] = None,
  table_kwargs: Mapping[str, Any] = MappingProxyType({}),
  column_kwargs: Mapping[str, Any] = MappingProxyType({}),
  compute_flops: bool = False,
  compute_vjp_flops: bool = False,
  **kwargs,
) -> Callable[..., str]:
  """Returns a function that creates a summary of the Module represented as a table.

  This function accepts most of the same arguments and internally calls
  `Module.init`, except that it returns a function of the form
  `(*args, **kwargs) -> str` where `*args` and `**kwargs` are passed to
  `method` (e.g. `__call__`) during the forward pass.

  `tabulate` uses `jax.eval_shape` under the hood to run the forward computation
  without consuming any FLOPs or allocating memory.

  Additional arguments can be passed into the `console_kwargs` argument, for
  example, `{'width': 120}`. For a full list of `console_kwargs` arguments, see:
  https://rich.readthedocs.io/en/stable/reference/console.html#rich.console.Console

  Example::

    >>> import flax.linen as nn
    >>> import jax, jax.numpy as jnp

    >>> class Foo(nn.Module):
    ...   @nn.compact
    ...   def __call__(self, x):
    ...     h = nn.Dense(4)(x)
    ...     return nn.Dense(2)(h)

    >>> x = jnp.ones((16, 9))
    >>> tabulate_fn = nn.tabulate(
    ...     Foo(), jax.random.key(0), compute_flops=True, compute_vjp_flops=True)

    >>> # print(tabulate_fn(x))

  This gives the following output::

                                           Foo Summary
    ┏━━━━━━━━━┳━━━━━━━━┳━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━┳━━━━━━━┳━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━┓
    ┃ path    ┃ module ┃ inputs        ┃ outputs       ┃ flops ┃ vjp_flops ┃ params          ┃
    ┡━━━━━━━━━╇━━━━━━━━╇━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━╇━━━━━━━╇━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━┩
    │         │ Foo    │ float32[16,9] │ float32[16,2] │ 1504  │ 4460      │                 │
    ├─────────┼────────┼───────────────┼───────────────┼───────┼───────────┼─────────────────┤
    │ Dense_0 │ Dense  │ float32[16,9] │ float32[16,4] │ 1216  │ 3620      │ bias:           │
    │         │        │               │               │       │           │ float32[4]      │
    │         │        │               │               │       │           │ kernel:         │
    │         │        │               │               │       │           │ float32[9,4]    │
    │         │        │               │               │       │           │                 │
    │         │        │               │               │       │           │ 40 (160 B)      │
    ├─────────┼────────┼───────────────┼───────────────┼───────┼───────────┼─────────────────┤
    │ Dense_1 │ Dense  │ float32[16,4] │ float32[16,2] │ 288   │ 840       │ bias:           │
    │         │        │               │               │       │           │ float32[2]      │
    │         │        │               │               │       │           │ kernel:         │
    │         │        │               │               │       │           │ float32[4,2]    │
    │         │        │               │               │       │           │                 │
    │         │        │               │               │       │           │ 10 (40 B)       │
    ├─────────┼────────┼───────────────┼───────────────┼───────┼───────────┼─────────────────┤
    │         │        │               │               │       │     Total │ 50 (200 B)      │
    └─────────┴────────┴───────────────┴───────────────┴───────┴───────────┴─────────────────┘

                                   Total Parameters: 50 (200 B)


  **Note**: rows order in the table does not represent execution order,
  instead it aligns with the order of keys in `variables` which are sorted
  alphabetically.

  **Note**: `vjp_flops` returns `0` if the module is not differentiable.

  Args:
    module: The module to tabulate.
    rngs: The rngs for the variable collections as passed to `Module.init`.
    depth: controls how many submodule deep the summary can go. By default its
      `None` which means no limit. If a submodule is not shown because of the
      depth limit, its parameter count and bytes will be added to the row of its
      first shown ancestor such that the sum of all rows always adds up to the
      total number of parameters of the Module.
    show_repeated: If `True`, repeated calls to the same module will be shown
      in the table, otherwise only the first call will be shown. Default is
      `False`.
    mutable: Can be bool, str, or list. Specifies which collections should be
      treated as mutable: ``bool``: all/no collections are mutable. ``str``: The
      name of a single mutable collection. ``list``: A list of names of mutable
      collections. By default all collections except 'intermediates' are
      mutable.
    console_kwargs: An optional dictionary with additional keyword arguments
      that are passed to `rich.console.Console` when rendering the table.
      Default arguments are `{'force_terminal': True, 'force_jupyter': False}`.
    table_kwargs: An optional dictionary with additional keyword arguments that
      are passed to `rich.table.Table` constructor.
    column_kwargs: An optional dictionary with additional keyword arguments that
      are passed to `rich.table.Table.add_column` when adding columns to the
      table.
    compute_flops: whether to include a `flops` column in the table listing the
      estimated FLOPs cost of each module forward pass. Does incur actual
      on-device computation / compilation / memory allocation, but still
      introduces overhead for large modules (e.g. extra 20 seconds for a
      Stable Diffusion's UNet, whereas otherwise tabulation would finish in 5
      seconds).
    compute_vjp_flops: whether to include a `vjp_flops` column in the table
      listing the estimated FLOPs cost of each module backward pass. Introduces
      a compute overhead of about 2-3X of `compute_flops`.
    **kwargs: Additional arguments passed to `Module.init`.

  Returns:
    A function that accepts the same `*args` and `**kwargs` of the forward pass
    (`method`) and returns a string with a tabular representation of the
    Modules.
  """
  # add non-default arguments to kwargs, this prevents some issue we overloading init
  # see: https://github.com/google/flax/issues/3299
  if mutable != DenyList('intermediates'):
    kwargs['mutable'] = mutable

  def _tabulate_fn(*fn_args, **fn_kwargs):
    table_fn = _get_module_table(
      module,
      depth=depth,
      show_repeated=show_repeated,
      compute_flops=compute_flops,
      compute_vjp_flops=compute_vjp_flops,
    )

    table = table_fn(rngs, *fn_args, **fn_kwargs, **kwargs)

    non_param_cols = [
      'path',
      'module',
      'inputs',
      'outputs',
    ]

    if compute_flops:
      non_param_cols.append('flops')
    if compute_vjp_flops:
      non_param_cols.append('vjp_flops')

    return _render_table(
      table, console_kwargs, table_kwargs, column_kwargs, non_param_cols
    )

  return _tabulate_fn


def _get_flops(fn, *args, **kwargs):
  e = jax.jit(fn).lower(*args, **kwargs)
  cost = e.cost_analysis()
  flops = int(cost['flops']) if 'flops' in cost else 0
  return flops


def _get_call_flops(
  c: module_lib._CallInfo,
  compute_flops: bool,
  compute_vjp_flops: bool,
) -> tuple[int, int]:
  """Return the FLOPs of executing the call `c` in the call stack.

  Does not perform actual computation / compilation / memory allocation, but
  still introduces overhead for large modules.

  Args:
    c: ``_CallInfo``.
    compute_flops: whether to compute forward pass FLOPs. Return `-1` otherwise.
    compute_vjp_flops: whether to compute backward pass FLOPs. Return `-1`
      otherwise.

  Returns:
    FLOPs of executing forward pass of `c`, and its VJP.
  """

  if not compute_flops and not compute_vjp_flops:
    return -1, -1

  rngs = jax.tree_map(
    lambda x: x.rng, c.rngs, is_leaf=lambda x: isinstance(x, LazyRng)
  )

  args = jax.tree_map(_from_value_representation, c.args)
  kwargs = jax.tree_map(_from_value_representation, c.kwargs)

  leaves, treedef = jax.tree_util.tree_flatten((args, kwargs))
  dynamic_leaves = []
  dynamic_idxs = []
  for i, arg in enumerate(leaves):
    if isinstance(arg, jax.ShapeDtypeStruct):
      dynamic_leaves.append(arg)
      dynamic_idxs.append(i)

  def _get_inputs(dynamic_leaves):
    new_leaves: list[Any] = leaves.copy()
    for i, arg in zip(dynamic_idxs, dynamic_leaves):
      new_leaves[i] = arg
    return treedef.unflatten(new_leaves)

  def init(rngs, dynamic_leaves):
    """`c.module.init` closed over static keyword arguments."""
    args, kwargs = _get_inputs(dynamic_leaves)
    return c.module.init(
      rngs,
      *args,
      method=c.method,
      mutable=c.mutable,
      **kwargs,
    )

  variables = jax.eval_shape(init, rngs, dynamic_leaves)

  def apply(variables, rngs, dynamic_leaves):
    """`c.module.apply` closed over static keyword arguments."""
    args, kwargs = _get_inputs(dynamic_leaves)
    return c.module.apply(
      variables,
      *args,
      rngs=rngs,
      method=c.method,
      mutable=c.mutable,
      **kwargs,
    )

  # Forward pass FLOPs
  if compute_flops:
    flops = _get_flops(apply, variables, rngs, dynamic_leaves)
  else:
    flops = -1

  if compute_vjp_flops:
    # Backward pass FLOPs
    def apply_vjp(variables, rngs, dynamic_leaves):
      """VJP of `c.module.apply` closed over static keyword arguments."""
      out, vjp_fn = jax.vjp(apply, variables, rngs, dynamic_leaves)
      return vjp_fn(out)

    vjp_flops = _get_flops(apply_vjp, variables, rngs, dynamic_leaves)
  else:
    vjp_flops = -1

  return flops, vjp_flops


def _get_module_table(
  module: module_lib.Module,
  depth: Optional[int],
  show_repeated: bool,
  compute_flops: bool,
  compute_vjp_flops: bool,
) -> Callable[..., Table]:
  """A function that takes a Module and returns function with the same signature
  as `init` but returns the Table representation of the Module."""

  def _get_table_fn(*args, **kwargs):
    with module_lib._tabulate_context():

      def _get_variables():
        return module.init(*args, **kwargs)

      variables = jax.eval_shape(_get_variables)
      calls = module_lib._context.call_info_stack[-1].calls
      calls.sort(key=lambda c: c.index)

    collections: Set[str] = set(variables.keys())
    rows = []
    all_paths: Set[Tuple[str, ...]] = set(call.path for call in calls)
    visited_paths: Set[Tuple[str, ...]] = set()

    for c in calls:
      call_depth = len(c.path)
      inputs = _process_inputs(c.args, c.kwargs)

      if c.path in visited_paths:
        if not show_repeated:
          continue
        module_vars = {}
        counted_vars = {}
      elif depth is not None:
        if call_depth > depth:
          continue
        module_vars, _ = _get_module_variables(c.path, variables, all_paths)
        if call_depth == depth:
          counted_vars = _get_path_variables(c.path, variables)
        else:
          counted_vars = module_vars
      else:
        module_vars, _ = _get_module_variables(c.path, variables, all_paths)
        counted_vars = module_vars

      visited_paths.add(c.path)
      rows.append(
        Row(
          c.path,
          type(c.module),
          c.method,
          inputs,
          c.outputs,
          module_vars,
          counted_vars,
          *_get_call_flops(c, compute_flops, compute_vjp_flops),
        )
      )

    return Table(module, tuple(collections), rows)

  return _get_table_fn


def _get_module_variables(
  path: Tuple[str, ...],
  variables: FrozenVariableDict,
  all_paths: Set[Tuple[str, ...]],
) -> Tuple[MutableVariableDict, Any]:
  """A function that takes a path and variables structure and returns a

  (module_variables, submodule_variables) tuple for that path.
  _get_module_variables
  uses the `all_paths` set to determine if a variable belongs to a submodule or
  not.
  """
  module_variables = _get_path_variables(path, variables)
  submodule_variables: Any = {collection: {} for collection in module_variables}
  all_keys = set(
    key for collection in module_variables.values() for key in collection
  )

  for key in all_keys:
    submodule_path = path + (key,)
    if submodule_path in all_paths:
      for collection in module_variables:
        if key in module_variables[collection]:
          submodule_variables[collection][key] = module_variables[
            collection
          ].pop(key)

  return module_variables, submodule_variables


def _get_path_variables(
  path: Tuple[str, ...], variables: FrozenVariableDict
) -> MutableVariableDict:
  """A function that takes a path and a variables structure and returns the
  variable structure at that path.
  """
  path_variables = {}

  for collection in variables:
    collection_variables = variables[collection]
    for name in path:
      if name not in collection_variables:
        collection_variables = None
        break
      collection_variables = collection_variables[name]

    if collection_variables is not None:
      path_variables[collection] = unfreeze(collection_variables)

  return path_variables


def _process_inputs(args, kwargs) -> Any:
  """A function that normalizes the representation of the ``args`` and
  ``kwargs`` for the ``inputs`` column.
  """
  if args and kwargs:
    input_values = (*args, kwargs)
  elif args and not kwargs:
    input_values = args[0] if len(args) == 1 else args
  elif kwargs and not args:
    input_values = kwargs
  else:
    input_values = ()

  return input_values


def _render_table(
  table: Table,
  console_extras: Optional[Mapping[str, Any]],
  table_kwargs: Mapping[str, Any],
  column_kwargs: Mapping[str, Any],
  non_params_cols: List[str],
) -> str:
  """A function that renders a Table to a string representation using rich."""
  console_kwargs = {'force_terminal': True, 'force_jupyter': False}
  if console_extras is not None:
    console_kwargs.update(console_extras)

  rich_table = rich.table.Table(
    show_header=True,
    show_lines=True,
    show_footer=True,
    title=f'{table.module.__class__.__name__} Summary',
    **table_kwargs,
  )

  for c in non_params_cols:
    rich_table.add_column(c, **column_kwargs)

  for col in table.collections:
    rich_table.add_column(col, **column_kwargs)

  for row in table:
    collections_size_repr = []

    for collection, size_bytes in row.size_and_bytes(table.collections).items():
      col_repr = ''

      if collection in row.module_variables:
        module_variables = _represent_tree(row.module_variables[collection])
        module_variables = _normalize_structure(module_variables)
        col_repr += _as_yaml_str(
          _summary_tree_map(_maybe_render, module_variables)
        )
        if col_repr:
          col_repr += '\n\n'

      col_repr += f'[bold]{_size_and_bytes_repr(*size_bytes)}[/bold]'
      collections_size_repr.append(col_repr)

    no_show_methods = {'__call__', '<lambda>'}
    path_repr = '/'.join(row.path)
    method_repr = (
      f' [dim]({row.method})[/dim]' if row.method not in no_show_methods else ''
    )
    rich_table.add_row(
      path_repr,
      row.module_type.__name__ + method_repr,
      *(
        _as_yaml_str(
          _summary_tree_map(
            _maybe_render, _normalize_structure(getattr(row, c))
          )
        )
        for c in non_params_cols[2:]
      ),
      *collections_size_repr,
    )

  # add footer with totals
  n_non_params_cols = len(non_params_cols)
  rich_table.columns[n_non_params_cols - 1].footer = rich.text.Text.from_markup(
    'Total', justify='right'
  )

  # get collection totals
  collection_total = {col: (0, 0) for col in table.collections}
  for row in table:
    for col, size_bytes in row.size_and_bytes(table.collections).items():
      collection_total[col] = (
        collection_total[col][0] + size_bytes[0],
        collection_total[col][1] + size_bytes[1],
      )

  # add totals to footer
  for i, col in enumerate(table.collections):
    rich_table.columns[n_non_params_cols + i].footer = _size_and_bytes_repr(
      *collection_total[col]
    )

  # add final totals to caption
  caption_totals = (0, 0)
  for size, num_bytes in collection_total.values():
    caption_totals = (
      caption_totals[0] + size,
      caption_totals[1] + num_bytes,
    )

  rich_table.caption_style = 'bold'
  rich_table.caption = (
    f'\nTotal Parameters: {_size_and_bytes_repr(*caption_totals)}'
  )

  return '\n' + _get_rich_repr(rich_table, console_kwargs) + '\n'


def _summary_tree_map(f, tree, *rest):
  return jax.tree_util.tree_map(f, tree, *rest, is_leaf=lambda x: x is None)


def _size_and_bytes_repr(size: int, num_bytes: int) -> str:
  if not size:
    return ''
  bytes_repr = _bytes_repr(num_bytes)
  return f'{size:,} [dim]({bytes_repr})[/dim]'


def _size_and_bytes(pytree: Any) -> Tuple[int, int]:
  leaves = jax.tree_util.tree_leaves(pytree)
  size = sum(x.size for x in leaves if hasattr(x, 'size'))
  num_bytes = sum(
    x.size * x.dtype.itemsize for x in leaves if hasattr(x, 'size')
  )
  return size, num_bytes


def _get_rich_repr(obj, console_kwargs):
  f = io.StringIO()
  console = rich.console.Console(file=f, **console_kwargs)
  console.print(obj)
  return f.getvalue()


def _as_yaml_str(value) -> str:
  if (hasattr(value, '__len__') and len(value) == 0) or value is None:
    return ''

  file = io.StringIO()
  yaml.safe_dump(
    value,
    file,
    default_flow_style=False,
    indent=2,
    sort_keys=False,
    explicit_end=False,
  )
  return file.getvalue().replace('\n...', '').replace("'", '').strip()


def _normalize_structure(obj):
  if isinstance(obj, _ValueRepresentation):
    return obj
  if isinstance(obj, (tuple, list)):
    return tuple(map(_normalize_structure, obj))
  elif isinstance(obj, Mapping):
    return {
      _normalize_structure(k): _normalize_structure(v) for k, v in obj.items()
    }
  elif dataclasses.is_dataclass(obj):
    return {
      f.name: _normalize_structure(getattr(obj, f.name))
      for f in dataclasses.fields(obj)
    }
  elif isinstance(obj, enum.Enum):
    # `yaml.safe_dump` does not support Enum key types so extract the underlying value
    return obj.value
  else:
    return obj


def _bytes_repr(num_bytes):
  count, units = (
    (f'{num_bytes / 1e9 :,.1f}', 'GB')
    if num_bytes > 1e9
    else (f'{num_bytes / 1e6 :,.1f}', 'MB')
    if num_bytes > 1e6
    else (f'{num_bytes / 1e3 :,.1f}', 'KB')
    if num_bytes > 1e3
    else (f'{num_bytes:,}', 'B')
  )

  return f'{count} {units}'


def _get_value_representation(x: Any) -> _ValueRepresentation:
  if isinstance(x, (int, float, bool, type(None))) or (
    isinstance(x, np.ndarray) and np.isscalar(x)
  ):
    return _ObjectRepresentation(x)
  elif isinstance(x, meta.Partitioned):
    return _PartitionedArrayRepresentation.from_partitioned(x)
  try:
    return _ArrayRepresentation.from_array(x)
  except:
    return _ObjectRepresentation(x)


def _from_value_representation(x: _ValueRepresentation) -> Any:
  if isinstance(x, _ArrayRepresentation):
    return jax.ShapeDtypeStruct(x.shape, x.dtype)

  elif isinstance(x, _PartitionedArrayRepresentation):
    return jax.ShapeDtypeStruct(
      x.array_representation.shape, x.array_representation.dtype
    )

  elif isinstance(x, _ObjectRepresentation):
    return x.obj

  raise TypeError(x, type(x))


def _represent_tree(x):
  """Returns a tree with the same structure as `x` but with each leaf replaced
  by a `_ValueRepresentation` object."""
  return jax.tree_util.tree_map(
    _get_value_representation,
    x,
    is_leaf=lambda x: x is None or isinstance(x, meta.Partitioned),
  )


def _maybe_render(x):
  return x.render() if hasattr(x, 'render') else repr(x)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Generic, TypeVar

import jax
import pytest

from flax import serialization
from flax.experimental import nnx


class TestPytree:
  def test_immutable_pytree(self):
    class Foo(nnx.Pytree):
      def __init__(self, y) -> None:
        self.x = 2
        self.y = nnx.Variable(y)

    pytree = Foo(y=3)

    leaves = jax.tree_util.tree_leaves(pytree)
    assert leaves == [3]

    pytree = jax.tree_map(lambda x: x * 2, pytree)
    assert pytree.x == 2
    assert pytree.y == 6

    pytree = pytree.replace(x=3)
    assert pytree.x == 3
    assert pytree.y == 6

    with pytest.raises(
      AttributeError, match='is immutable, trying to update field'
    ):
      pytree.x = 4

  def test_immutable_pytree_dataclass(self):
    @nnx.dataclass(frozen=True)
    class Foo(nnx.Pytree):
      y: int = nnx.treenode_field()
      x: int = nnx.field(default=2)

    pytree = Foo(y=3)

    leaves = jax.tree_util.tree_leaves(pytree)
    assert leaves == [3]

    pytree = jax.tree_map(lambda x: x * 2, pytree)
    assert pytree.x == 2
    assert pytree.y == 6

    pytree = pytree.replace(x=3)
    assert pytree.x == 3
    assert pytree.y == 6

    with pytest.raises(AttributeError, match='cannot assign to field'):
      pytree.x = 4

  def test_jit(self):
    @nnx.dataclass
    class Foo(nnx.Pytree):
      a: int = nnx.treenode_field()
      b: int = nnx.field()

    module = Foo(a=1, b=2)

    @jax.jit
    def f(m: Foo):
      return m.a + m.b

    assert f(module) == 3

  def test_flax_serialization(self):
    class Bar(nnx.Pytree):
      def __init__(self, a, b):
        self.a = a
        self.b = nnx.Variable(b)

    @nnx.dataclass
    class Foo(nnx.Pytree):
      bar: Bar
      c: int = nnx.treenode_field()
      d: int = nnx.field()

    foo: Foo = Foo(bar=Bar(a=1, b=2), c=3, d=4)

    state_dict = serialization.to_state_dict(foo)

    assert state_dict == {
      'bar': {
        'b': 2,
      },
      'c': 3,
    }

    state_dict['bar']['b'] = 5

    foo = serialization.from_state_dict(foo, state_dict)

    assert foo.bar.b == 5

    del state_dict['bar']['b']

    with pytest.raises(ValueError, match='Missing field'):
      serialization.from_state_dict(foo, state_dict)

    state_dict['bar']['b'] = 5

    # add unknown field
    state_dict['x'] = 6

    with pytest.raises(ValueError, match='Unknown field'):
      serialization.from_state_dict(foo, state_dict)

  def test_generics(self):
    T = TypeVar('T')

    class MyClass(nnx.Pytree, Generic[T]):
      def __init__(self, x: T):
        self.x = x

    MyClass[int]

  def test_key_paths(self):
    @nnx.dataclass
    class Bar(nnx.Pytree):
      a: int = nnx.treenode_field(default=1)
      b: int = nnx.field(default=2)

    @nnx.dataclass
    class Foo(nnx.Pytree):
      x: int = nnx.treenode_field(default=3)
      y: int = nnx.field(default=4)
      z: Bar = nnx.treenode_field(default_factory=Bar)

    foo = Foo()

    path_values, treedef = jax.tree_util.tree_flatten_with_path(foo)
    path_values = [(list(map(str, path)), value) for path, value in path_values]

    assert path_values[0] == (['.x', '.value'], 3)
    assert path_values[1] == (['.z', '.value', '.a', '.value'], 1)

  def test_replace_unknown_fields_error(self):
    class Foo(nnx.Pytree):
      pass

    with pytest.raises(ValueError, match='Trying to replace unknown fields'):
      Foo().replace(y=1)

  def test_dataclass_inheritance(self):
    @nnx.dataclass
    class A(nnx.Pytree):
      a: int = nnx.treenode_field(default=1)
      b: int = nnx.field(default=2)

    @nnx.dataclass
    class B(A):
      c: int = nnx.treenode_field(default=3)

    pytree = B()
    leaves = jax.tree_util.tree_leaves(pytree)
    assert leaves == [1, 3]

  def test_pytree_with_new(self):
    class A(nnx.Pytree):
      def __init__(self, a):
        self.a = a

      def __new__(cls, a):
        return super().__new__(cls)

    pytree = A(a=1)

    pytree = jax.tree_map(lambda x: x * 2, pytree)

  def test_deterministic_order(self):
    class A(nnx.Pytree):
      def __init__(self, order: bool):
        if order:
          self.a = 1
          self.b = 2
        else:
          self.b = 2
          self.a = 1

    p1 = A(order=True)
    p2 = A(order=False)

    leaves1 = jax.tree_util.tree_leaves(p1)
    leaves2 = jax.tree_util.tree_leaves(p2)

    assert leaves1 == leaves2


class TestMutablePytree:
  def test_pytree(self):
    class Foo(nnx.Pytree, mutable=True):
      def __init__(self, y) -> None:
        self.x = 2
        self.y = nnx.Variable(y)

    pytree = Foo(y=3)

    leaves = jax.tree_util.tree_leaves(pytree)
    assert leaves == [3]

    pytree = jax.tree_map(lambda x: x * 2, pytree)
    assert pytree.x == 2
    assert pytree.y == 6

    pytree = pytree.replace(x=3)
    assert pytree.x == 3
    assert pytree.y == 6

    # test mutation
    pytree.x = 4
    assert pytree.x == 4

  def test_no_new_fields_after_init(self):
    class Foo(nnx.Pytree, mutable=True):
      def __init__(self, x):
        self.x = nnx.Variable(x)

    foo = Foo(x=1)
    foo.x = 2

    with pytest.raises(AttributeError, match=r'Cannot add new fields to'):
      foo.y = 2

  def test_pytree_dataclass(self):
    @nnx.dataclass
    class Foo(nnx.Pytree, mutable=True):
      y: int = nnx.treenode_field()
      x: int = nnx.field(default=2)

    pytree: Foo = Foo(y=3)

    leaves = jax.tree_util.tree_leaves(pytree)
    assert leaves == [3]

    pytree = jax.tree_map(lambda x: x * 2, pytree)
    assert pytree.x == 2
    assert pytree.y == 6

    pytree = pytree.replace(x=3)
    assert pytree.x == 3
    assert pytree.y == 6

    # test mutation
    pytree.x = 4
    assert pytree.x == 4



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Combinators of modules, such as a Sequential."""

from typing import Any, Callable, Dict, Sequence

from flax.linen.module import Module


class Sequential(Module):
  """Applies a linear chain of Modules.

  Meant to be used only for the simple case of fusing together callables where
  the input of a particular module/op is the output of the previous one.

  Modules will be applied in the order that they are passed in the constructor.

  The ``__call__`` method of Sequential accepts any input and forwards it to the
  first module it contains. It chains the output sequentially to the input of
  the next module and returns the output of the final module.

  Example usage::

    >>> import flax.linen as nn

    >>> class Foo(nn.Module):
    ...   @nn.compact
    ...   def __call__(self, x):
    ...     return nn.Sequential([nn.Dense(4),
    ...                           nn.relu,
    ...                           nn.Dense(2),
    ...                           nn.log_softmax])(x)

  This combinator supports also layers that return multiple outputs if returned
  as a tuple or a dictionary. If the output of a layer is a ``tuple`` it will be
  expanded as ``*args`` in the next layer, if its a ``dict`` it
  will be expanded as ``**kwargs``.

  Example usage::

    >>> class CrossAttentionBlock(nn.Module):
    ...   num_heads: int = 2
    ...   qkv_features: int = 16
    ...
    ...   @nn.compact
    ...   def __call__(self, query, key_value):
    ...     output = nn.MultiHeadDotProductAttention(
    ...       num_heads=self.num_heads, qkv_features=self.qkv_features)(query,
    ...                                                                 key_value)
    ...     output = nn.Dense(self.qkv_features)(output)
    ...     return dict(query=output, key_value=key_value)  # also works for tuples

    >>> from typing import Sequence
    >>> class CrossAttentionNetwork(nn.Module):
    ...   num_layers: Sequence[int]
    ...
    ...   @nn.compact
    ...   def __call__(self, x):
    ...     return nn.Sequential([CrossAttentionBlock() for _ in
    ...                           range(self.num_layers)])(query, key_value)
  """

  layers: Sequence[Callable[..., Any]]

  def __post_init__(self):
    if not isinstance(self.layers, Sequence):
      raise ValueError(
        f"'layers' must be a sequence, got '{type(self.layers).__name__}'."
      )
    super().__post_init__()

  def __call__(self, *args, **kwargs):
    if not self.layers:
      raise ValueError(f'Empty Sequential module {self.name}.')

    outputs = self.layers[0](*args, **kwargs)
    for layer in self.layers[1:]:
      if isinstance(outputs, tuple):
        outputs = layer(*outputs)
      elif isinstance(outputs, Dict):
        outputs = layer(**outputs)
      else:
        outputs = layer(outputs)
    return outputs



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Attention core modules for Flax."""

from __future__ import annotations

import functools
from typing import Any, Callable, Optional, Tuple, overload

import jax
import jax.numpy as jnp
from jax import lax, random

from flax.experimental import nnx
from flax.experimental.nnx.nnx import rnglib
from flax.experimental.nnx.nnx import flaglib
from flax.experimental.nnx.nnx.module import Module, first_from
from flax.experimental.nnx.nnx.nn import initializers
from flax.experimental.nnx.nnx.nn.dtypes import promote_dtype
from flax.experimental.nnx.nnx.nn.linear import (
  DotGeneralT,
  LinearGeneral,
  PrecisionLike,
  default_kernel_init,
)
from flax.experimental.nnx.nnx.nn.normalization import LayerNorm

Array = jax.Array
Shape = Tuple[int, ...]
Dtype = Any


def dot_product_attention_weights(
  query: Array,
  key: Array,
  bias: Optional[Array] = None,
  mask: Optional[Array] = None,
  broadcast_dropout: bool = True,
  dropout_rng: Optional[Array] = None,
  dropout_rate: float = 0.0,
  deterministic: bool = False,
  dtype: Optional[Dtype] = None,
  precision: PrecisionLike = None,
  module: Optional[Module] = None,
):
  """Computes dot-product attention weights given query and key.

  Used by :func:`dot_product_attention`, which is what you'll most likely use.
  But if you want access to the attention weights for introspection, then
  you can directly call this function and call einsum yourself.

  Args:
    query: queries for calculating attention with shape of `[batch..., q_length,
      num_heads, qk_depth_per_head]`.
    key: keys for calculating attention with shape of `[batch..., kv_length,
      num_heads, qk_depth_per_head]`.
    bias: bias for the attention weights. This should be broadcastable to the
      shape `[batch..., num_heads, q_length, kv_length]`. This can be used for
      incorporating causal masks, padding masks, proximity bias, etc.
    mask: mask for the attention weights. This should be broadcastable to the
      shape `[batch..., num_heads, q_length, kv_length]`. This can be used for
      incorporating causal masks. Attention weights are masked out if their
      corresponding mask value is `False`.
    broadcast_dropout: bool: use a broadcasted dropout along batch dims.
    dropout_rng: JAX PRNGKey: to be used for dropout
    dropout_rate: dropout rate
    deterministic: bool, deterministic or not (to apply dropout)
    dtype: the dtype of the computation (default: infer from inputs and params)
    precision: numerical precision of the computation see `jax.lax.Precision`
      for details.
    module: the Module that will sow the attention weights into the
      ``nnx.Intermediate`` collection. If ``module`` is None, the attention
      weights will not be sowed.

  Returns:
    Output of shape `[batch..., num_heads, q_length, kv_length]`.
  """
  query, key = promote_dtype(query, key, dtype=dtype)
  dtype = query.dtype

  assert query.ndim == key.ndim, 'q, k must have same rank.'
  assert query.shape[:-3] == key.shape[:-3], 'q, k batch dims must match.'
  assert query.shape[-2] == key.shape[-2], 'q, k num_heads must match.'
  assert query.shape[-1] == key.shape[-1], 'q, k depths must match.'

  # calculate attention matrix
  depth = query.shape[-1]
  query = query / jnp.sqrt(depth).astype(dtype)
  # attn weight shape is (batch..., num_heads, q_length, kv_length)
  attn_weights = jnp.einsum(
    '...qhd,...khd->...hqk', query, key, precision=precision
  )

  # apply attention bias: masking, dropout, proximity bias, etc.
  if bias is not None:
    attn_weights = attn_weights + bias
  # apply attention mask
  if mask is not None:
    big_neg = jnp.finfo(dtype).min
    attn_weights = jnp.where(mask, attn_weights, big_neg)

  # normalize the attention weights
  attn_weights = jax.nn.softmax(attn_weights).astype(dtype)

  if module:
    module.sow(nnx.Intermediate, 'attention_weights', attn_weights)

  # apply attention dropout
  if not deterministic and dropout_rate > 0.0:
    keep_prob = 1.0 - dropout_rate
    if broadcast_dropout:
      # dropout is broadcast across the batch + head dimensions
      dropout_shape = tuple([1] * (key.ndim - 2)) + attn_weights.shape[-2:]
      keep = random.bernoulli(dropout_rng, keep_prob, dropout_shape)  # type: ignore
    else:
      keep = random.bernoulli(dropout_rng, keep_prob, attn_weights.shape)  # type: ignore
    multiplier = keep.astype(dtype) / jnp.asarray(keep_prob, dtype=dtype)
    attn_weights = attn_weights * multiplier

  return attn_weights


def dot_product_attention(
  query: Array,
  key: Array,
  value: Array,
  bias: Optional[Array] = None,
  mask: Optional[Array] = None,
  broadcast_dropout: bool = True,
  dropout_rng: Optional[Array] = None,
  dropout_rate: float = 0.0,
  deterministic: bool = False,
  dtype: Optional[Dtype] = None,
  precision: PrecisionLike = None,
  module: Optional[Module] = None,
):
  """Computes dot-product attention given query, key, and value.

  This is the core function for applying attention based on
  https://arxiv.org/abs/1706.03762. It calculates the attention weights given
  query and key and combines the values using the attention weights.

  Note: query, key, value needn't have any batch dimensions.

  Args:
    query: queries for calculating attention with shape of `[batch..., q_length,
      num_heads, qk_depth_per_head]`.
    key: keys for calculating attention with shape of `[batch..., kv_length,
      num_heads, qk_depth_per_head]`.
    value: values to be used in attention with shape of `[batch..., kv_length,
      num_heads, v_depth_per_head]`.
    bias: bias for the attention weights. This should be broadcastable to the
      shape `[batch..., num_heads, q_length, kv_length]`. This can be used for
      incorporating causal masks, padding masks, proximity bias, etc.
    mask: mask for the attention weights. This should be broadcastable to the
      shape `[batch..., num_heads, q_length, kv_length]`. This can be used for
      incorporating causal masks. Attention weights are masked out if their
      corresponding mask value is `False`.
    broadcast_dropout: bool: use a broadcasted dropout along batch dims.
    dropout_rng: JAX PRNGKey: to be used for dropout
    dropout_rate: dropout rate
    deterministic: bool, deterministic or not (to apply dropout)
    dtype: the dtype of the computation (default: infer from inputs)
    precision: numerical precision of the computation see `jax.lax.Precision`
      for details.
    module: the Module that will sow the attention weights into the
      ``nnx.Intermediate`` collection. If ``module`` is None, the attention
      weights will not be sowed.

  Returns:
    Output of shape `[batch..., q_length, num_heads, v_depth_per_head]`.
  """
  query, key, value = promote_dtype(query, key, value, dtype=dtype)
  dtype = query.dtype
  assert key.ndim == query.ndim == value.ndim, 'q, k, v must have same rank.'
  assert (
    query.shape[:-3] == key.shape[:-3] == value.shape[:-3]
  ), 'q, k, v batch dims must match.'
  assert (
    query.shape[-2] == key.shape[-2] == value.shape[-2]
  ), 'q, k, v num_heads must match.'
  assert key.shape[-3] == value.shape[-3], 'k, v lengths must match.'

  # compute attention weights
  attn_weights = dot_product_attention_weights(
    query,
    key,
    bias,
    mask,
    broadcast_dropout,
    dropout_rng,
    dropout_rate,
    deterministic,
    dtype,
    precision,
    module,
  )

  # return weighted sum over values for each query position
  return jnp.einsum(
    '...hqk,...khd->...qhd', attn_weights, value, precision=precision
  )


class MultiHeadAttention(Module):
  """Multi-head attention.

  Example usage::

    >>> import flax.linen as nn
    >>> import jax

    >>> layer = nn.MultiHeadAttention(num_heads=8, qkv_features=16)
    >>> key1, key2, key3, key4, key5, key6 = jax.random.split(jax.random.key(0), 6)
    >>> shape = (4, 3, 2, 5)
    >>> q, k, v = jax.random.uniform(key1, shape), jax.random.uniform(key2, shape), jax.random.uniform(key3, shape)
    >>> variables = layer.init(jax.random.key(0), q)

    >>> # different inputs for inputs_q, inputs_k and inputs_v
    >>> out = layer.apply(variables, q, k, v)
    >>> # equivalent to layer.apply(variables, inputs_q=q, inputs_k=k, inputs_v=k)
    >>> out = layer.apply(variables, q, k)
    >>> # equivalent to layer.apply(variables, inputs_q=q, inputs_k=q) and layer.apply(variables, inputs_q=q, inputs_k=q, inputs_v=q)
    >>> out = layer.apply(variables, q)

    >>> attention_kwargs = dict(
    ...     num_heads=8,
    ...     qkv_features=16,
    ...     kernel_init=nn.initializers.ones,
    ...     bias_init=nn.initializers.zeros,
    ...     dropout_rate=0.5,
    ...     deterministic=False,
    ...     )
    >>> class Module(nn.Module):
    ...   attention_kwargs: dict
    ...
    ...   @nn.compact
    ...   def __call__(self, x, dropout_rng=None):
    ...     out1 = nn.MultiHeadAttention(**self.attention_kwargs)(x, dropout_rng=dropout_rng)
    ...     out2 = nn.MultiHeadAttention(**self.attention_kwargs)(x, dropout_rng=dropout_rng)
    ...     return out1, out2
    >>> module = Module(attention_kwargs)
    >>> variables = module.init({'params': key1, 'dropout': key2}, q)

    >>> # out1 and out2 are different.
    >>> out1, out2 = module.apply(variables, q, rngs={'dropout': key3})
    >>> # out3 and out4 are different.
    >>> # out1 and out3 are different. out2 and out4 are different.
    >>> out3, out4 = module.apply(variables, q, rngs={'dropout': key4})
    >>> # out1 and out2 are the same.
    >>> out1, out2 = module.apply(variables, q, dropout_rng=key5)
    >>> # out1 and out2 are the same as out3 and out4.
    >>> # providing a `dropout_rng` arg will take precedence over the `rngs` arg in `.apply`
    >>> out3, out4 = module.apply(variables, q, rngs={'dropout': key6}, dropout_rng=key5)

  Attributes:
    num_heads: number of attention heads. Features (i.e. inputs_q.shape[-1])
      should be divisible by the number of heads.
    dtype: the dtype of the computation (default: infer from inputs and params)
    param_dtype: the dtype passed to parameter initializers (default: float32)
    qkv_features: dimension of the key, query, and value.
    out_features: dimension of the last projection
    broadcast_dropout: bool: use a broadcasted dropout along batch dims.
    dropout_rate: dropout rate
    deterministic: if false, the attention weight is masked randomly using
      dropout, whereas if true, the attention weights are deterministic.
    precision: numerical precision of the computation see `jax.lax.Precision`
      for details.
    kernel_init: initializer for the kernel of the Dense layers.
    bias_init: initializer for the bias of the Dense layers.
    use_bias: bool: whether pointwise QKVO dense transforms use bias.
    attention_fn: dot_product_attention or compatible function. Accepts query,
      key, value, and returns output of shape `[bs, dim1, dim2, ..., dimN,,
      num_heads, value_channels]``
    decode: whether to prepare and use an autoregressive cache.
    normalize_qk: should QK normalization be applied (arxiv.org/abs/2302.05442).
  """

  def __init__(
    self,
    num_heads: int,
    in_features: int,
    qkv_features: int | None = None,
    out_features: int | None = None,
    *,
    dtype: Dtype | None = None,
    param_dtype: Dtype = jnp.float32,
    broadcast_dropout: bool = True,
    dropout_rate: float = 0.0,
    deterministic: bool | None = None,
    precision: PrecisionLike = None,
    kernel_init: initializers.Initializer = default_kernel_init,
    bias_init: initializers.Initializer = initializers.zeros_init(),
    use_bias: bool = True,
    attention_fn: Callable[..., Array] = dot_product_attention,
    decode: bool | None = None,
    normalize_qk: bool = False,
    # Deprecated, will be removed.
    qkv_dot_general: DotGeneralT | None = None,
    out_dot_general: DotGeneralT | None = None,
    qkv_dot_general_cls: Any = None,
    out_dot_general_cls: Any = None,
    rngs: rnglib.Rngs,
  ):
    self.num_heads = num_heads
    self.in_features = in_features
    self.qkv_features = (
      qkv_features if qkv_features is not None else in_features
    )
    self.out_features = (
      out_features if out_features is not None else in_features
    )
    self.dtype = dtype
    self.param_dtype = param_dtype
    self.broadcast_dropout = broadcast_dropout
    self.dropout_rate = dropout_rate
    self.deterministic = deterministic
    self.precision = precision
    self.kernel_init = kernel_init
    self.bias_init = bias_init
    self.use_bias = use_bias
    self.attention_fn = attention_fn
    self.decode = decode
    self.normalize_qk = normalize_qk
    self.qkv_dot_general = qkv_dot_general
    self.out_dot_general = out_dot_general
    self.qkv_dot_general_cls = qkv_dot_general_cls
    self.out_dot_general_cls = out_dot_general_cls

    if self.qkv_features % self.num_heads != 0:
      raise ValueError(
        f'Memory dimension ({self.qkv_features}) must be divisible by '
        f"'num_heads' heads ({self.num_heads})."
      )

    self.head_dim = self.qkv_features // self.num_heads

    linear_general = functools.partial(
      LinearGeneral,
      in_features=self.in_features,
      out_features=(self.num_heads, self.head_dim),
      dtype=self.dtype,
      param_dtype=self.param_dtype,
      kernel_init=self.kernel_init,
      bias_init=self.bias_init,
      use_bias=self.use_bias,
      precision=self.precision,
      dot_general=self.qkv_dot_general,
      dot_general_cls=self.qkv_dot_general_cls,
    )
    # project inputs_q to multi-headed q/k/v
    # dimensions are then [batch..., length, n_heads, n_features_per_head]
    self.query = linear_general(rngs=rngs)
    self.key = linear_general(rngs=rngs)
    self.value = linear_general(rngs=rngs)

    if self.normalize_qk:
      # Normalizing query and key projections stabilizes training with higher
      # LR. See ViT-22B paper http://arxiv.org/abs/2302.05442 for analysis.
      self.query_ln = LayerNorm(self.head_dim, use_bias=False, rngs=rngs)
      self.key_ln = LayerNorm(self.head_dim, use_bias=False, rngs=rngs)
    else:
      self.query_ln = None
      self.key_ln = None

    self.out = LinearGeneral(
      in_features=(self.num_heads, self.head_dim),
      out_features=self.out_features,
      axis=(-2, -1),
      kernel_init=self.kernel_init,
      bias_init=self.bias_init,
      use_bias=self.use_bias,
      dtype=self.dtype,
      param_dtype=self.param_dtype,
      precision=self.precision,
      dot_general=self.out_dot_general,
      dot_general_cls=self.out_dot_general_cls,
      rngs=rngs,
    )

  @overload
  def __call__(
    self,
    inputs_q: Array,
    inputs_k: Optional[Array] = None,
    inputs_v: Optional[Array] = None,
    *,
    mask: Optional[Array] = None,
    deterministic: Optional[bool] = None,
    dropout_rng: Optional[Array] = None,
    rngs: rnglib.Rngs | None = None,
    sow_weights: bool = False,
    decode: bool | None = None,
  ):
    ...

  @overload
  def __call__(
    self,
    inputs_q: Array,
    *,
    inputs_kv: Array | None = None,
    mask: Array | None = None,
    deterministic: bool | None = None,
    dropout_rng: Array | None = None,
    rngs: rnglib.Rngs | None = None,
    sow_weights: bool = False,
    decode: bool | None = None,
  ):
    ...

  def __call__(
    self,
    inputs_q: Array,
    inputs_k: Array | None = None,
    inputs_v: Array | None = None,
    *,
    mask: Array | None = None,
    deterministic: bool | None = None,
    rngs: rnglib.Rngs | None = None,
    sow_weights: bool = False,
    decode: bool | None = None,
  ):
    """Applies multi-head dot product attention on the input data.

    Projects the inputs into multi-headed query, key, and value vectors,
    applies dot-product attention and project the results to an output vector.

    If both inputs_k and inputs_v are None, they will both copy the value of
    inputs_q (self attention).
    If only inputs_v is None, it will copy the value of inputs_k.

    Args:
      inputs_q: input queries of shape `[batch_sizes..., length, features]`.
      inputs_k: key of shape `[batch_sizes..., length, features]`. If None,
        inputs_k will copy the value of inputs_q.
      inputs_v: values of shape `[batch_sizes..., length, features]`. If None,
        inputs_v will copy the value of inputs_k.
      mask: attention mask of shape `[batch_sizes..., num_heads, query_length,
        key/value_length]`. Attention weights are masked out if their
        corresponding mask value is `False`.
      deterministic: if false, the attention weight is masked randomly using
        dropout, whereas if true, the attention weights are deterministic.
      rngs: container for random number generators to generate the dropout
        mask when `deterministic` is False. The `rngs` container should have a
        `dropout` key.
      sow_weights: if ``True``, the attention weights are sowed into the
        'intermediates' collection.

    Returns:
      output of shape `[batch_sizes..., length, features]`.
    """

    if inputs_k is None:
      if inputs_v is not None:
        raise ValueError(
          '`inputs_k` cannot be None if `inputs_v` is not None. '
          'To have both `inputs_k` and `inputs_v` be the same value, pass in the '
          'value to `inputs_k` and leave `inputs_v` as None.'
        )
      inputs_k = inputs_q
    if inputs_v is None:
      inputs_v = inputs_k

    if inputs_q.shape[-1] != self.in_features:
      raise ValueError(
        f'Incompatible input dimension, got {inputs_q.shape[-1]} '
        f'but module expects {self.in_features}.'
      )

    query = self.query(inputs_q)
    key = self.key(inputs_k)
    value = self.value(inputs_v)

    if self.normalize_qk:
      assert self.query_ln is not None and self.key_ln is not None
      # Normalizing query and key projections stabilizes training with higher
      # LR. See ViT-22B paper http://arxiv.org/abs/2302.05442 for analysis.
      query = self.query_ln(query)
      key = self.key_ln(key)

    # During fast autoregressive decoding, we feed one position at a time,
    # and cache the keys and values step by step.
    decode = first_from(
      decode,
      self.decode,
      flaglib.flags.get('decode'),
      error_msg="""No `decode` argument was provided to MultiHeadAttention
        as either a __call__ argument, class attribute, or nnx.flag.""",
    )

    if decode:
      (
        *batch_dims,
        max_length,
        num_heads,
        depth_per_head,
      ) = self.cached_key.shape
      # shape check of cached keys against query input
      expected_shape = tuple(batch_dims) + (1, num_heads, depth_per_head)
      if expected_shape != query.shape:
        raise ValueError(
          'Autoregressive cache shape error, '
          'expected query shape %s instead got %s.'
          % (expected_shape, query.shape)
        )
      # update key, value caches with our new 1d spatial slices
      cur_index = self.cache_index
      zero = jnp.array(0, dtype=lax.dtype(cur_index.dtype))
      indices = (zero,) * len(batch_dims) + (cur_index, zero, zero)
      key = lax.dynamic_update_slice(self.cached_key, key, indices)
      value = lax.dynamic_update_slice(self.cached_value, value, indices)
      self.cached_key = key
      self.cached_value = value
      self.cache_index += 1
      # causal mask for cached decoder self-attention:
      # our single query position should only attend to those key
      # positions that have already been generated and cached,
      # not the remaining zero elements.
      mask = combine_masks(
        mask,
        jnp.broadcast_to(
          jnp.arange(max_length) <= cur_index,
          tuple(batch_dims) + (1, 1, max_length),
        ),
      )

    if (
      self.dropout_rate > 0.0
    ):  # Require `deterministic` only if using dropout.
      deterministic = first_from(
        deterministic,
        self.deterministic,
        flaglib.flags.get('deterministic'),
        error_msg="""No `deterministic` argument was provided to MultiHeadAttention
          as either a __call__ argument, class attribute, or nnx.flag.""",
      )
      if not deterministic:
        if rngs is None:
          raise ValueError(
            "'rngs' must be provided if 'dropout_rng' is not given."
          )
        dropout_rng = rngs.dropout()
      else:
        dropout_rng = None
    else:
      deterministic = True
      dropout_rng = None

    # apply attention
    x = self.attention_fn(
      query,
      key,
      value,
      mask=mask,
      dropout_rng=dropout_rng,
      dropout_rate=self.dropout_rate,
      broadcast_dropout=self.broadcast_dropout,
      deterministic=deterministic,
      dtype=self.dtype,
      precision=self.precision,
      module=self if sow_weights else None,
    )
    # back to the original inputs dimensions
    out = self.out(x)
    return out

  def init_cache(self, input_shape: Shape, dtype: Dtype = jnp.float32):
    """Initializes cache for fast autoregressive decoding."""
    cache_shape = (*input_shape[:-1], self.num_heads, self.head_dim)
    self.cached_key = nnx.Cache(jnp.zeros(cache_shape, dtype))
    self.cached_value = nnx.Cache(jnp.zeros(cache_shape, dtype))
    self.cache_index = nnx.Cache(jnp.array(0, dtype=jnp.int32))


# mask-making utility functions


def make_attention_mask(
  query_input: Array,
  key_input: Array,
  pairwise_fn: Callable[..., Any] = jnp.multiply,
  extra_batch_dims: int = 0,
  dtype: Dtype = jnp.float32,
):
  """Mask-making helper for attention weights.

  In case of 1d inputs (i.e., `[batch..., len_q]`, `[batch..., len_kv]`, the
  attention weights will be `[batch..., heads, len_q, len_kv]` and this
  function will produce `[batch..., 1, len_q, len_kv]`.

  Args:
    query_input: a batched, flat input of query_length size
    key_input: a batched, flat input of key_length size
    pairwise_fn: broadcasting elementwise comparison function
    extra_batch_dims: number of extra batch dims to add singleton axes for, none
      by default
    dtype: mask return dtype

  Returns:
    A `[batch..., 1, len_q, len_kv]` shaped mask for 1d attention.
  """
  mask = pairwise_fn(
    jnp.expand_dims(query_input, axis=-1), jnp.expand_dims(key_input, axis=-2)
  )
  mask = jnp.expand_dims(mask, axis=-3)
  mask = jnp.expand_dims(mask, axis=tuple(range(extra_batch_dims)))
  return mask.astype(dtype)


def make_causal_mask(
  x: Array, extra_batch_dims: int = 0, dtype: Dtype = jnp.float32
) -> Array:
  """Make a causal mask for self-attention.

  In case of 1d inputs (i.e., `[batch..., len]`, the self-attention weights
  will be `[batch..., heads, len, len]` and this function will produce a
  causal mask of shape `[batch..., 1, len, len]`.

  Args:
    x: input array of shape `[batch..., len]`
    extra_batch_dims: number of batch dims to add singleton axes for, none by
      default
    dtype: mask return dtype

  Returns:
    A `[batch..., 1, len, len]` shaped causal mask for 1d attention.
  """
  idxs = jnp.broadcast_to(jnp.arange(x.shape[-1], dtype=jnp.int32), x.shape)
  return make_attention_mask(
    idxs,
    idxs,
    jnp.greater_equal,
    extra_batch_dims=extra_batch_dims,
    dtype=dtype,
  )


def combine_masks(*masks: Optional[Array], dtype: Dtype = jnp.float32) -> Array:
  """Combine attention masks.

  Args:
    *masks: set of attention mask arguments to combine, some can be None.
    dtype: dtype for the returned mask.

  Returns:
    Combined mask, reduced by logical and, returns None if no masks given.
  """
  masks_list = [m for m in masks if m is not None]
  if not masks_list:
    return None
  assert all(
    map(lambda x: x.ndim == masks_list[0].ndim, masks_list)
  ), f'masks must have same rank: {tuple(map(lambda x: x.ndim, masks_list))}'
  mask, *other_masks = masks_list
  for other_mask in other_masks:
    mask = jnp.logical_and(mask, other_mask)
  return mask.astype(dtype)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for working with jit and partitioned models.

This module introduces ``axis_rules``, ``logical_to_mesh_axes``,
``logical_to_mesh``, ``with_logical_constraint`` for appyling jit sharding
constraints in terms of "logical named axes" rather than jit's default mesh
axes.

Additionally the ``LogicallyPartitioned`` metadata wrapper is defined as
well as the initializer function wrapper ``with_logical_partitioning``for
introducing logical axis metadata into a model's variables.
"""

import collections
import contextlib
import dataclasses
import enum
import functools
import threading
from typing import Any, Callable, List, Optional, Sequence, Tuple, Union

import jax
from jax import lax
from jax.experimental import maps

from flax import struct
from flax.core import meta

# Real types and dummy aliases for documentation
LogicalRules = Sequence[Tuple[str, Union[str, Tuple[str], None]]]
Array = Any  # pylint: disable=invalid-name
ArrayPytree = Any  # pylint: disable=invalid-name
LogicalPartitionSpec = Any  # pylint: disable=invalid-name
LogicalPartitionSpecPytree = Any  # pylint: disable=invalid-name
PartitionSpecPytree = Any  # pylint: disable=invalid-name


# Dynamic Axis Mapping Context
# ------------------------------------------------------------------------------


@dataclasses.dataclass
class _AxisRules(threading.local):
  """Dynamic logical axis to mesh axis binding context."""

  rules: LogicalRules = ()


# Global axis binding context.
_axis_rules = _AxisRules()


def set_logical_axis_rules(rules: LogicalRules):
  """Sets the global logical axis to mesh axis binding."""
  _axis_rules.rules = rules


def get_logical_axis_rules() -> LogicalRules:
  """Returns the global logical axis to mesh axis binding."""
  return _axis_rules.rules


@contextlib.contextmanager
def logical_axis_rules(rules: LogicalRules):
  """Context manager for setting the logical to mesh axis bindings."""
  old_rules = _axis_rules.rules
  try:
    _axis_rules.rules = rules
    yield
  finally:
    _axis_rules.rules = old_rules


class _UnassignedAxis:
  """Sentinel class for unassigned logical axis name."""

  def __repr__(self):
    return 'UnassignedAxis'

  def __bool__(self):
    return False


_unassigned_axis = _UnassignedAxis()


def _mesh_assignment_free(new_assignment, existing_assignments):
  """Determines if a given mesh axis has already been assigned."""
  new = set(jax.tree_util.tree_leaves(new_assignment))
  existing = set(jax.tree_util.tree_leaves(existing_assignments))
  if existing.intersection(new):
    return False
  return True


def _logical_to_mesh_axes(
  array_dim_names: Optional[Sequence[Optional[str]]],
  rules: Optional[LogicalRules] = None,
) -> Optional[List[Union[_UnassignedAxis, None, str, Tuple[str]]]]:
  """Same as logical_to_mesh_axes, but doesn't fill in _unassigned_axis."""
  if array_dim_names is None:
    return None
  if rules is None:
    rules = _axis_rules.rules
  axis_name_counts = collections.Counter(array_dim_names)
  dups = tuple(
    k for k, v in axis_name_counts.items() if v > 1 and k is not None
  )
  if dups:
    raise ValueError(
      f'Unsupported: Dimensions {dups} occur more than once in array names.'
    )
  if not isinstance(rules, (tuple, list)):
    raise ValueError('Unknown axis rule specification type.')
  # We assign mesh axes using a priority based ruleset over logical axis names.
  result: List[Union[_UnassignedAxis, None, str, Tuple[str]]]
  result = [_unassigned_axis] * len(array_dim_names)
  for rule_model_name, rule_mesh_names in rules:
    if rule_model_name in array_dim_names:
      pos = array_dim_names.index(rule_model_name)
      if (
        _mesh_assignment_free(rule_mesh_names, result)
        and result[pos] == _unassigned_axis
      ):
        result[pos] = rule_mesh_names
  return result


def logical_to_mesh_axes(
  array_dim_names: Optional[Sequence[Optional[str]]],
  rules: Optional[LogicalRules] = None,
) -> Optional[jax.sharding.PartitionSpec]:
  """Compute layout for an array.

  The rules are in order of precedence, and consist of pairs:
  ``(ArrayDimensionName, MeshDimensionName)``, meaning that the given array
  dimension (if present and unused) should be sharded across the given
  mesh dimension (if present and unused).

  A Layout of an Array is expressed as a tuple with one element for each
  dimension in the Array. The element is either None, or is the name of a
  mesh-dimension, meaning that this dimension of the array is sharded across
  this dimension of the mesh.

  For example, given an array with::

    array_dim_names = ('batch', 'length', 'heads', 'features')

  and the layout rules are::

    rules = (('batch', 'X'),
             ('features', 'X'),
             ('heads', 'Y'),
             ('batch', 'Z'))

  then this function will return::

    PartitionSpec('X', None, 'Y', None)

  Args:
    array_dim_names: Tuple of array dimension names or None.
    rules: Optional logical to mesh rules override.  Defaults to using the
      rules defined in the dynamic context set from the ``axis_rules`` function.

  Returns:
    PartitionSpec for the parameter.
  """
  result = _logical_to_mesh_axes(array_dim_names, rules)
  if result is None:
    return None
  # We default to None - ie unsharded along the dimension.
  result = [None if x is _unassigned_axis else x for x in result]
  return jax.sharding.PartitionSpec(*result)


def logical_to_mesh(tree: Any, rules: Optional[LogicalRules] = None) -> Any:
  """Applies logical_to_mesh_axes to pytrees of logical PartitionSpecs."""
  return jax.tree_map(
    lambda x: logical_to_mesh_axes(x, rules),
    tree,
    is_leaf=lambda x: isinstance(x, jax.sharding.PartitionSpec),
  )


def logical_to_mesh_sharding(
  tree: Any,
  mesh: jax.sharding.Mesh,
  rules: Optional[LogicalRules] = None,
) -> Any:
  """Convert pytrees of logical PartitionSpecs to shardings."""
  return jax.tree_map(
    lambda x: jax.sharding.NamedSharding(mesh, x),
    logical_to_mesh(tree, rules),
    is_leaf=lambda x: isinstance(x, jax.sharding.PartitionSpec),
  )


def _global_mesh_defined() -> bool:
  """Checks if global xmap/jit mesh resource environment is defined."""
  maps_env = maps.thread_resources.env
  return maps_env.physical_mesh.devices.shape != ()  # pylint: disable=g-explicit-bool-comparison


class RulesFallback(enum.Enum):
  """How a sharding constraint should behave when no matching rule is found."""

  AXIS_IS_UNSHARDED = 'axis_is_unsharded'
  RAISE_ERROR = 'raise_error'
  NO_CONSTRAINT = 'no_constraint'


def _with_sharding_constraint(
  x: Array,
  axis_resources: Optional[jax.sharding.PartitionSpec],
  mesh: Optional[jax.sharding.Mesh] = None,
):
  """Wrapper for lax.with_sharding_constraint, no-op on cpu or outside jit."""
  if jax.devices()[0].platform == 'cpu' or (
    not _global_mesh_defined() and mesh is None
  ):
    return x
  else:
    if mesh is not None and axis_resources is not None:
      sharding = jax.sharding.NamedSharding(mesh, axis_resources)
      return lax.with_sharding_constraint(x, sharding)
    return lax.with_sharding_constraint(x, axis_resources)


def _with_sharding_constraint_one_fallback(
  axis_resources: LogicalPartitionSpec,
  x: Array,
  fallback: RulesFallback = RulesFallback.AXIS_IS_UNSHARDED,
  rules: Optional[LogicalRules] = None,
  mesh: Optional[jax.sharding.Mesh] = None,
):
  """Either imposes a sharding constraint or applies fallback."""
  mesh_axes = _logical_to_mesh_axes(axis_resources, rules)
  if mesh_axes is None:
    return _with_sharding_constraint(x, None, mesh=mesh)

  if fallback == RulesFallback.AXIS_IS_UNSHARDED:
    mesh_axes = [None if x is _unassigned_axis else x for x in mesh_axes]
  else:
    if any(x is _unassigned_axis for x in mesh_axes):
      if fallback == RulesFallback.RAISE_ERROR:
        raise ValueError(f'Axis names {axis_resources} did not match a rule')
      else:
        return x
  return _with_sharding_constraint(
    x, jax.sharding.PartitionSpec(*mesh_axes), mesh=mesh
  )


def _is_logical_spec(x):
  return x is None or (
    isinstance(x, tuple) and all(isinstance(e, str) or e is None for e in x)
  )


def with_logical_constraint(
  x: ArrayPytree,
  logical_axis_resources: LogicalPartitionSpecPytree,
  rules: Optional[LogicalRules] = None,
  mesh: Optional[jax.sharding.Mesh] = None,
  fallback: RulesFallback = RulesFallback.AXIS_IS_UNSHARDED,
):
  """Version of jit's with_sharding_constraint that uses logical axis names."""
  # If no axis binding is set, this is a no-op.
  if rules is None:
    rules = _axis_rules.rules
  if not rules or logical_axis_resources is None:
    return x
  # Translate logical names to mesh assignments.
  return jax.tree_util.tree_map(
    functools.partial(
      _with_sharding_constraint_one_fallback,
      fallback=fallback,
      rules=rules,
      mesh=mesh,
    ),
    logical_axis_resources,
    x,
    is_leaf=_is_logical_spec,
  )


# Logical Partitioning Axis Metadata
# ------------------------------------------------------------------------------


class LogicallyPartitioned(meta.Partitioned):
  rules: Optional[LogicalRules] = struct.field(default=None, pytree_node=False)

  def unbox(self, apply_constraint=True) -> Any:
    """Returns the wrapped value with the partitioning constraint applied."""
    if apply_constraint and (_global_mesh_defined() or self.mesh is not None):
      return with_logical_constraint(
        self.value,
        self.get_partition_spec(),
        rules=self.rules,
        mesh=self.mesh,
      )
    else:
      return self.value


def with_logical_partitioning(
  fn: Callable[..., Any],
  names: meta.LogicalNames,
  mesh: Optional[jax.sharding.Mesh] = None,
  rules: Optional[LogicalRules] = None,
) -> Callable[..., LogicallyPartitioned]:
  """Wraps a function's return value with LogicallyPartitioned.

  Example::

    >>> import flax.linen as nn
    >>> kernel_init = nn.with_logical_partitioning(
    ...     nn.initializers.lecun_normal, (None, "data"))
    >>> partitioned_dense = nn.Dense(features=3, kernel_init=kernel_init)

  Args:
    fn: The function to be wrapped. Typically this is an initializer.
    names: The logical axis passed to ``LogicallyPartitioned``.
    mesh: The mesh to use for the partitioning. If None, the global mesh
      resource is used if available.
    rules: Optional logical to mesh rules use. If None, the global rules
      are used if available.
  Returns:
    A function wrapping ``fn`` that will return an instance of
    ``LogicallyPartitioned``.
  """

  @functools.wraps(fn)
  def wrapper(*args, **kwargs):
    return LogicallyPartitioned(
      fn(*args, **kwargs), names, mesh=mesh, rules=rules
    )  # pytype: disable=wrong-keyword-args

  return wrapper



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Any, Optional

import jax
from absl.testing import absltest
from jax import numpy as jnp
from jax import random

from flax import struct
from flax.core import Array, init, nn, unfreeze


@dataclass
class Dense:
  features: int
  bias: bool = True
  kernel_init: Any = nn.linear.default_kernel_init
  bias_init: Any = nn.initializers.zeros_init()

  def __call__(self, scope, x):
    kernel = scope.param(
      'kernel', self.kernel_init, (x.shape[-1], self.features)
    )
    y = x @ kernel
    if self.bias:
      bias = scope.param('bias', self.bias_init, (self.features,))
      y += bias.reshape((1,) * (y.ndim - 1) + (-1,))
    return y


@struct.dataclass
class ExplicitDense:
  kernel: Array
  bias: Optional[Array]

  # a fully explicit "scope free" version
  @staticmethod
  def create(
    rng,
    in_size,
    out_size,
    bias=True,
    kernel_init=nn.linear.default_kernel_init,
    bias_init=nn.initializers.zeros_init(),
  ):
    k1, k2 = random.split(rng, 2)
    kernel = kernel_init(k1, (in_size, out_size))
    if bias:
      bias = bias_init(k2, (out_size,))
    else:
      bias = None
    return ExplicitDense(kernel, bias)

  # a semi-explicit version where a scope is used to create explicit params
  @staticmethod
  def create_in_scope(
    scope,
    in_size,
    out_size,
    bias=True,
    kernel_init=nn.linear.default_kernel_init,
    bias_init=nn.initializers.zeros_init(),
  ):
    kernel = scope.param('kernel', kernel_init, (in_size, out_size))
    if bias:
      bias = scope.param('bias', bias_init, (out_size,))
    else:
      bias = None
    return ExplicitDense(kernel, bias)

  def __call__(self, x):
    y = x @ self.kernel
    if self.bias is not None:
      y += self.bias.reshape((1,) * (y.ndim - 1) + (-1,))
    return y


def explicit_mlp(scope, x, sizes=(3, 1)):
  for i, size in enumerate(sizes):
    dense = scope.param(f'dense_{i}', ExplicitDense.create, x.shape[-1], size)
    x = dense(x)
    if i + 1 < len(sizes):
      x = nn.relu(x)
  return x


def semi_explicit_mlp(scope, x, sizes=(3, 1)):
  for i, size in enumerate(sizes):
    dense = scope.child(ExplicitDense.create_in_scope, prefix='dense_')(
      x.shape[-1], size
    )
    x = dense(x)
    if i + 1 < len(sizes):
      x = nn.relu(x)
  return x


class DenseTest(absltest.TestCase):
  def test_dense(self):
    model = Dense(features=4)
    x = jnp.ones((1, 3))
    y, variables = init(model)(random.key(0), x)
    param_shapes = unfreeze(
      jax.tree_util.tree_map(jnp.shape, variables['params'])
    )
    self.assertEqual(y.shape, (1, 4))
    self.assertEqual(
      param_shapes,
      {
        'kernel': (3, 4),
        'bias': (4,),
      },
    )

  def test_explicit_dense(self):
    x = jnp.ones((1, 3))
    y, variables = init(explicit_mlp)(random.key(0), x)
    param_shapes = unfreeze(
      jax.tree_util.tree_map(jnp.shape, variables['params'])
    )
    self.assertEqual(y.shape, (1, 4))
    self.assertEqual(
      param_shapes,
      {
        'kernel': (3, 4),
        'bias': (4,),
      },
    )

  def test_explicit_dense(self):
    x = jnp.ones((1, 4))
    y, variables = init(explicit_mlp)(random.key(0), x)
    param_shapes = unfreeze(
      jax.tree_util.tree_map(jnp.shape, variables['params'])
    )
    self.assertEqual(y.shape, (1, 1))
    self.assertEqual(
      param_shapes,
      {
        'dense_0': ExplicitDense((4, 3), (3,)),
        'dense_1': ExplicitDense((3, 1), (1,)),
      },
    )

  def test_semi_explicit_dense(self):
    x = jnp.ones((1, 4))
    y, variables = init(semi_explicit_mlp)(random.key(0), x)
    param_shapes = unfreeze(
      jax.tree_util.tree_map(jnp.shape, variables['params'])
    )
    self.assertEqual(y.shape, (1, 1))
    self.assertEqual(
      param_shapes,
      {
        'dense_0': {'kernel': (4, 3), 'bias': (3,)},
        'dense_1': {'kernel': (3, 1), 'bias': (1,)},
      },
    )


if __name__ == '__main__':
  absltest.main()
